(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* Implementation of the SCENARIO signature for the Bogeholz case
   study. Mathematically, this file implements a structure for the
   signature in the file bogeholz.d.
*)

structure BogeholzScenario :> SCENARIO = 
struct

structure Db = SQLiteDb

val db_name: Db.dbName = "bogeholz.db"


val declarations: Declaration.decls = 
    let val declFileName = "bogeholz.d"
	val ds = Option.valOf (DeclParser.parse_decls_file declFileName)
	val _ = Declaration.check_decls ds
    in
	ds
    end


local
    open Fol
    open Time

    exception BadInput
    exception CannotOpenConnection
    exception NotAttribute
    exception Impossible

    (* Convert n days to Time.time *)
    fun fromDays (n: int): Time.time = 
	let val secs = (LargeInt.fromInt n) * (LargeInt.fromInt 86400)
	in
	    Time.fromSeconds secs
	end
	
    (* Not required in the signature, but we need it to implement hasattr
       and inrole predicates *) 
    structure Sql = SQLFn (Db)

in
    fun eval_term c tl = 
	case (c, tl) of
	    ("seven_year", _) => tInt (365 * 7) 
	    (* Relative time is represented as integers in days *)
	  | ("one_year", _) => tInt (365 * 1)
	  | ("thirty_day", _) => tInt 30
	  | ("plus", [tDate (t: Time.time), tInt n]) =>	tDate (t + (fromDays n))
	  | _ => (print ("In BogeholzScenario.eval_term: Term " ^ 
			 (print_term (tApp(c,tl))) ^ 
			 " not understood\n");
		  raise BadInput)


    (* Return the super attributes of a given attribute *)
    fun super_attrs (t: term) = 
	case t of
	    tApp ("rec", [k]) => []
	  | tApp ("rec/static", [k]) => [tApp ("rec", [k])]
	  | tApp ("rec/dynamic", [k,u]) => [tApp ("rec", [k])]
	  | tApp ("rec/static/personal", [k]) => 
	    [tApp ("rec/static", [k]), tApp ("rec", [k])]
	  | tApp ("rec/static/financial", [k]) => 
	    [tApp ("rec/static", [k]), tApp ("rec", [k])]
	  | tApp ("rec/static/emergency", [k]) => 
	    [tApp ("rec/static", [k]), tApp ("rec", [k])]
	  | tApp ("rec/static/pastmedical", [k]) => 
	    [tApp ("rec/static", [k]), tApp ("rec", [k])]
	  | tApp ("rec/dynamic/careplan", [k,u]) => 
	    [tApp ("rec/dynamic", [k,u]), tApp ("rec", [k])]
	  | tApp ("rec/dynamic/progressnotes", [k,u]) => 
	    [tApp ("rec/dynamic", [k,u]), tApp ("rec", [k])]
	  | tApp ("rec/dynamic/medicalrecord", [k,k',u]) => 
	    [tApp ("rec/dynamic", [k,u]), tApp ("rec", [k])]
	  | tApp ("rec/dynamic/medicalrecord/rehab", [k,k',u]) => 
	    [tApp ("rec/dynamic/medicalrecord", [k,k',u]), 
	     tApp ("rec/dynamic", [k,u]), tApp ("rec", [k])]
	  | tApp ("rec/dynamic/medicalrecord/private", [k,k',u]) => 
	    [tApp ("rec/dynamic/medicalrecord", [k,k',u]), 
	     tApp ("rec/dynamic", [k,u]), tApp ("rec", [k])]
	  | tApp ("rec/dynamic/medicalrecord/psycho", [k,k',u]) => 
	    [tApp ("rec/dynamic/medicalrecord", [k,k',u]), 
	     tApp ("rec/dynamic", [k,u]), tApp ("rec", [k])]
	  | _ => (print ("Badly-formed attribute: " ^ 
			 (print_term t) ^ "\n");
		  raise NotAttribute)

	    

    fun resolve_eval_pred connection p tl sc = 
	case (p, tl) of
	    ("inrole", [k, r, tDate T]) =>
	    (* k, r, u are input mode here *)

	    (* We need to check whether k is in role r at time T or
	       not. Our strategy is: 

               1. Find the set U_s = {u_s | beginrole (k, r, u_s)}
               2. Find the largest u_s \in U_s such that 0 <= u_s <= T
                  (no such u_s => inrole(k,r,u) does NOT hold)
               3. Find the set U_e = {u_e | endrole (k, r, u_e)}
               4. Find a u_e \in U_e such that u_s < u_e <= T
                  (such u_e exists => inrole(k,r,u) does NOT hold, else it does)
	     *)

	    let val u_s = ref (Time.zeroTime) 
		(* u_s will eventually contain the largest u satisfying beginrole(k,r,u) and 0 <= u <= T *) 
			  
		val exists_s = ref (false)
		(* Will be true iff we find such a u_s *)
			       
		val _ = Sql.resolve_db_pred declarations connection "beginrole" [k, r, tVar "_"]
					    (fn Solution.NOMORE => ()
					      | Solution.NEXT [_,_,tDate u] => 
						if (u >= !u_s andalso u <= T) then (u_s := u; exists_s := true) else ()
					      | Solution.NEXT tl => raise Impossible
					    )
	    in
		if (!exists_s = false) 
		then
		    (* There wasn't a beginrole (k,r,...) event before T *)
		    sc (Solution.NOMORE)
		else
                    (* !u_s is the largest u satisfying beginrole(k,r,u) and 0 <= u <= T *)
		    let 			 
			local
			    exception Found
			in
			
			(* exists_e is true iff there is a u_e such that endrole(k,r,u_e) and T >= u_e > u_s *)
			val exists_e = 
			    (Sql.resolve_db_pred declarations connection "endrole" [k, r, tVar "_"]
						 (fn Solution.NOMORE => ()
						   | Solution.NEXT [_,_,tDate u] => 
						     if (u <= T andalso u > !u_s) then raise Found else ()
						   | Solution.NEXT tl => raise Impossible
						 );
			     false) handle Found => true
			end
			
		    in
			if (exists_e) 
			then (* The role ended before or at T *)
			    sc (Solution.NOMORE)
			else
			    (sc (Solution.NEXT [k, r, tDate T]);
			     sc (Solution.NOMORE))
		    end
	    end
	    
	  | ("hasattr_ext", [t1, _]) => 
	    (* t1 is input mode here *)
	    let 
		(* Call the success continuation with all solutions in list *)
		fun call_all sc [] = raise Impossible
		  | call_all sc [attr] = sc (Solution.NEXT [t1,attr])
		  | call_all sc (attr :: al) =
		    (sc (Solution.NEXT [t1, attr]); call_all sc al)
	    in
		Sql.resolve_db_pred declarations connection "hasattr" [t1, tVar "_"]
				    (fn Solution.NOMORE => sc Solution.NOMORE
				      | Solution.NEXT [_, attr] =>
					let val al = super_attrs attr
					in
					    call_all sc (attr :: al)
					end
				      | Solution.NEXT tl =>
					(print ("Unexpected output from hasattr table: " ^
						(print_atom ("hasattr", tl)) ^ "\n");
					 raise Impossible
					)
				    )	 
	    end

 
	  | ("eq_action", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("eq_role", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("eq_prin", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("eq_time", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("lessthan", [tDate t1, tDate t2]) => 
            (* t1 and t2 are both in input mode here *)
	    if (t1 <= t2) 
	    then
		(sc (Solution.NEXT [tDate t1, tDate t2]); 
		 sc (Solution.NOMORE))
	    else 
		sc (Solution.NOMORE)

	  | _ => (print ("In BogeholzScenario.resolve_eval_pred: Atom " ^ 
			 (print_atom (p,tl)) ^ 
			 " not understood\n");
		  raise BadInput)

	
val subj_pred_list_true = ref []
val subj_pred_list_false = ref []

fun isFinal p tl = 
    (* Ideally, this should check the time argument of tl and compare
       to clock time; here we will assume that databases are complete,
       so everything is always final *)
    true

end (* local *)

end
