/*
Copyright (c) <2010>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>


#include <iostream>

#ifdef USE_DB
#include <sqlite3.h>
#endif

#include "charbuf.hpp"
#include "bogeholz-consts.hpp"
#include "ded-sim.hpp"
#include "boundedbuffer.hpp"
#include "NativeTime.hpp"

using namespace std;



//////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////

#ifdef USE_DB
sqlite3 * pDb; // Global pointer to database
static const char dbname[]  = "bogeholz.db";
#endif


int eventCount = 0;
// A counter to order events at the same time; passed in the priority
// field

int residentCounter = 0;
int prCounter = 0;
int doctorCounter = 0;
int nurseCounter = 0;
int hcwCounter = 0;
int adminstaffCounter = 0;
int volunteerCounter = 0;
int managerCounter = 0;

int physioCounter = 0;
int infoCounter = 0;
int lawCounter = 0;

int currResidents[nResident];
int doctorOf[nResident]; // doctorOf[i] = k => resident i has doctor k
int currPrs[nResident];
bool incapable[nResident];
int currDoctors[nDoctor]; 
int currNurses[nNurse];
int currHcws[nHcw];
int currAdminStaffs[nAdminStaff];
int currVolunteers[nVolunteer];
int currManager;

boundedbuffer<int> * activeRecords[nResident];



////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

char * intTimeToString (int t)
{
  static char out[256];
  NativeTime nt(t, true);

  nt.toString (out);
  return out;
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////


int run_no_result(const charbuf & cmd_buf)
{
#ifdef USE_DB
  sqlite3_stmt * pStmt;
  
  if (sqlite3_prepare_v2(pDb, cmd_buf.getHead(), -1, &pStmt, NULL) != SQLITE_OK) {
      cerr << "Error preparing the statement:\n<<<<<<<<\n" << cmd_buf << "\n>>>>>>>>\n";
      return 0;
  }
  
  if (sqlite3_step(pStmt) != SQLITE_DONE) {
    cerr << "Error running the statement:\n<<<<<<<<\n" << cmd_buf << "\n>>>>>>>>\n";
    sqlite3_finalize(pStmt);
    return 0;
  }

  sqlite3_finalize(pStmt);
#else
  cout << cmd_buf.getHead() << endl;
#endif
  return 1;
}




///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////



class ReplaceResident: public AbstractEvent
// Replace a resident with a new one
{
public:

  int res_no; // 0 <= res_no < nResident. Which resident are we replacing?

  ReplaceResident(int t, int p, int r_no)
  {
    time = t;
    priority = p;
    res_no = r_no;
  }

  void fire();

};


class ReplaceManager: public AbstractEvent
// Replace the manager with a new one
{
public:
  
  ReplaceManager(int t, int p) 
  {
    time = t;
    priority = p;
  }

  void fire();
};

class ReplaceNurse: public AbstractEvent
// Replace a nurse with a new one
{
public:
  int nurse_no; // 0 <= nurse_no < nNurse. Which nurse are we replacing?

  ReplaceNurse(int t, int p, int n_no) 
  {
    time = t;
    priority = p;
    nurse_no = n_no;
  }

  void fire();
};

class ReplaceHcw: public AbstractEvent
// Replace a hcw with a new one
{
public:
  int hcw_no; // 0 <= hcw_no < nHcw. Which hcw are we replacing?

  ReplaceHcw(int t, int p, int h_no) 
  {
    time = t;
    priority = p;
    hcw_no = h_no;
  }

  void fire();
};



class ReplaceAdminStaff: public AbstractEvent
// Replace a adminstaff with a new one
{
public:
  int adminstaff_no; // 0 <= adminstaff_no < nAdminStaff. Which adminstaff are we replacing?

  ReplaceAdminStaff(int t, int p, int a_no) 
  {
    time = t;
    priority = p;
    adminstaff_no = a_no;
  }

  void fire();
};


class ReplaceVolunteer: public AbstractEvent
// Replace a volunteer with a new one
{
public:
  int volunteer_no; // 0 <= volunteer_no < nVolunteer. Which volunteer are we replacing?

  ReplaceVolunteer(int t, int p, int v_no) 
  {
    time = t;
    priority = p;
    volunteer_no = v_no;
  }

  void fire();
};


class ReplaceDoctor: public AbstractEvent
// Replace a doctor with a new one
{
public:
  int doc_no; // 0 <= doc_no < nDoctor. Which doctor are we replacing?

  ReplaceDoctor(int t, int p, int d_no)
  {
    time = t;
    priority = p;
    doc_no = d_no;
  }

  void fire();
};


class SwitchDoctor: public AbstractEvent
// Switch the doctor of a resident
{
public:
  int res_no; // Resident whose doctor needs to be switched
  int res_id; // Id of resident. If currResidents[res_no] != res_id,
	      // then the event should not fire
  int doc_id; // Id of current doctor. If
	      // currDoctors[doctorOf[res_no]] != doc_id, then the
	      // event should not fire

  SwitchDoctor(int t, int p, int r_no, int r_id, int d_id) 
  {
    time = t;
    priority = p;
    res_no = r_no;
    res_id = r_id;
    doc_id = d_id;
  }

  void fire();
};



class MakeResidentIncapable: public AbstractEvent
// Make a resident incapable
{
public:
  int res_no; // Resident to be made incapable
  int res_id; // Id of resident. If currResidents[res_no] != res_id,
	      // then the event should not fire

  MakeResidentIncapable(int t, int p, int r_no, int r_id) 
  {
    time = t;
    priority = p;
    res_no = r_no;
    res_id = r_id;
  }

  void fire();
};



class MakeResidentCapable: public AbstractEvent
// Make a resident capable
{
public:
  int res_no; // Resident to be made capable
  int res_id; // Id of resident. If currResidents[res_no] != res_id,
	      // then the event should not fire

  MakeResidentCapable(int t, int p, int r_no, int r_id) 
  {
    time = t;
    priority = p;
    res_no = r_no;
    res_id = r_id;
  }

  void fire();
};



class ReadResident: public AbstractEvent
// Resident reads his record (legitimately)
{
public:
  int res_no; // Resident who wants to read
  int res_id; // Id of resident. If currResidents[res_no] != res_id,
	      // then the event should not fire

  ReadResident(int t, int p, int r_no, int r_id) 
  {
    time = t;
    priority = p;
    res_no = r_no;
    res_id = r_id;
  }

  void fire();
};



class CheatResident: public AbstractEvent
// Resident reads random record (cheat)
{
public:
  int res_no; // Resident who wants to read
  int res_id; // Id of resident. If currResidents[res_no] != res_id,
	      // then the event should not fire

  CheatResident(int t, int p, int r_no, int r_id) 
  {
    time = t;
    priority = p;
    res_no = r_no;
    res_id = r_id;
  }

  void fire();
};


class CheatManager: public AbstractEvent
// Manager reads random record (cheat)
{
public:
  int man_id; // Id of manager. If currManager != man_id, then this
	      // event should not fire

  CheatManager(int t, int p, int m_id) 
  {
    time = t;
    priority = p;
    man_id = m_id;
  }

  void fire();
};


class CheatNurse: public AbstractEvent
// Nurse reads random record (cheat)
{
public:
  int nurse_no; // Nurse who wants to read
  int nurse_id; 
  // Id of nurse. If currNurses[nurse_no] != nurse_id, then the event
  // should not fire

  CheatNurse(int t, int p, int n_no, int n_id) 
  {
    time = t;
    priority = p;
    nurse_no = n_no;
    nurse_id = n_id;
  }

  void fire();
};


class CheatHcw: public AbstractEvent
// Hcw reads random record (cheat)
{
public:
  int hcw_no; // Hcw who wants to read
  int hcw_id; 
  // Id of hcw. If currHcws[hcw_no] != hcw_id, then the event
  // should not fire

  CheatHcw(int t, int p, int h_no, int h_id) 
  {
    time = t;
    priority = p;
    hcw_no = h_no;
    hcw_id = h_id;
  }

  void fire();
};


class CheatAdminStaff: public AbstractEvent
// AdminStaff reads random record (cheat)
{
public:
  int adminstaff_no; // AdminStaff who wants to read
  int adminstaff_id; 
  // Id of adminstaff. If currAdminStaffs[adminstaff_no] != adminstaff_id, then the event
  // should not fire

  CheatAdminStaff(int t, int p, int a_no, int a_id) 
  {
    time = t;
    priority = p;
    adminstaff_no = a_no;
    adminstaff_id = a_id;
  }

  void fire();
};



class CheatVolunteer: public AbstractEvent
// Volunteer reads random record (cheat)
{
public:
  int volunteer_no; // Volunteer who wants to read
  int volunteer_id; 
  // Id of volunteer. If currVolunteers[volunteer_no] != volunteer_id, then the event
  // should not fire

  CheatVolunteer(int t, int p, int v_no, int v_id) 
  {
    time = t;
    priority = p;
    volunteer_no = v_no;
    volunteer_id = v_id;
  }

  void fire();
};



class AddCareplan: public AbstractEvent
// Add a new careplan for a resident (manager adds the careplan)
{
public:
  int res_no; // Resident whose careplan must be added
  int res_id; // Id of resident. If currResidents[res_no] != res_id,
	      // then the event should not fire

  AddCareplan(int t, int p, int r_no, int r_id) 
  {
    time = t;
    priority = p;
    res_no = r_no;
    res_id = r_id;
  }

  void fire();
};


class DoctorVisit: public AbstractEvent
// The event of a doctor making a visit
{
public:
  int doc_no; // Doctor who is visiting
  
  DoctorVisit(int t, int p, int d_no) 
  {
    time = t;
    priority = p;
    doc_no = d_no;
  }

  void fire();
};


class PhysioVisit: public AbstractEvent
// The event of a physiotherapist making a visit
{
public:
  int physio_id; // Physiotherapist who is visiting. Note that this is
		 // not a serial number; it is the id itself
		 // (physiotherapists have no serial numbers)

  int res_no; // Resident being visited
  
  PhysioVisit(int t, int p, int p_id, int r_no) 
  {
    time = t;
    priority = p;
    physio_id = p_id;
    res_no = r_no;
  }

  void fire();
};


class NurseVisit: public AbstractEvent
// A nurse makes her periodic visit
{
public:
  int nurse_no; // Nurse making the visit

  NurseVisit(int t, int p, int n_no)
  {
    time = t;
    priority = p;
    nurse_no = n_no;
  }

  void fire();
};


class HcwVisit: public AbstractEvent
// A hwc makes her periodic visit
{
public:
  int hcw_no; // Hcw making the visit

  HcwVisit(int t, int p, int h_no)
  {
    time = t;
    priority = p;
    hcw_no = h_no;
  }

  void fire();
};


class VolunteerAdd: public AbstractEvent
// Volunteer adds a progress note to a random resident
{
public:
  int volunteer_no; // Volunteer making the update

  VolunteerAdd(int t, int p, int v_no)
  {
    time = t;
    priority = p;
    volunteer_no = v_no;
  }

  void fire();
};


class ExResidentRequest: public AbstractEvent
// Random ex-resident makes a request
{
public:

  ExResidentRequest(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};


class ManagerResponse: public AbstractEvent
// Manager responds to a request from an ex-resident
{
public:
  int res_id; // Id of resident to whom response is sent

  ManagerResponse(int t, int p, int r_id)
  {
    time = t;
    priority = p;
    res_id = r_id;
  }

  void fire();
};


class AbuseDisclosure: public AbstractEvent
// Manager believes that there is an abuse case and discloses to a law officer
{
public:
  
  AbuseDisclosure(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};


///////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////

void ReplaceManager::fire ()
{
  charbuf man_name(256);
  charbuf stmt(1024);

  if (currManager != -1) {
    // A current manager exists, so we should end his role
    man_name << "(str2prin \"_m" << currManager << "\")";
    
    stmt << "INSERT INTO endrole VALUES ("
	 << "'" << man_name.getHead() << "',"
	 << "'manager', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
  }
  
  // Set new current manager
  currManager = managerCounter++;

  // New manager name
  man_name.reset();
  man_name << "(str2prin \"_m" << currManager << "\")";

  // Insert beginrole (man_name, manager, u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << man_name.getHead() << "',"
       << "'manager', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);
    
  // Add confidentiality agreement (but forget to do it with a small
  // probability)
  bool forgetConfAgreement = ((uniform_double() < forgetConf) ? true: false);

  if (!forgetConfAgreement) {
    stmt.reset();
    stmt << "INSERT INTO signed_conf VALUES (" 
	 << "'" << man_name.getHead() << "', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);    
  }

  // Add event to have manager read random record (cheat)
  int cheattime = this -> time + exp_int (meanGapManagerCheat);
  addEvent (new CheatManager(cheattime, eventCount++, currManager)); 


  // Add next event to replace this manager
  int replacetime = this -> time + exp_int (meanManagerLife);
  addEvent (new ReplaceManager(replacetime, eventCount++));
}


void ReplaceNurse::fire ()
{
  charbuf nurse_name(256);
  charbuf stmt(1024);

  if (currNurses[this -> nurse_no] != -1) {
    // A current nurse exists, so we should end his role
    nurse_name << "(str2prin \"_n" << currNurses[this -> nurse_no] << "\")";
    
    stmt << "INSERT INTO endrole VALUES ("
	 << "'" << nurse_name.getHead() << "',"
	 << "'nurse', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
  }
  
  // Set new current nurse
  currNurses[this -> nurse_no] = nurseCounter++;

  // New nurse name
  nurse_name.reset();
  nurse_name << "(str2prin \"_n" << currNurses[this -> nurse_no] << "\")";

  // Insert beginrole (nurse_name, nurse, u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << nurse_name.getHead() << "',"
       << "'nurse', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);
    
  // Add confidentiality agreement (but forget to do it with a small
  // probability)
  bool forgetConfAgreement = ((uniform_double() < forgetConf) ? true: false);

  if (!forgetConfAgreement) {
    stmt.reset();
    stmt << "INSERT INTO signed_conf VALUES (" 
	 << "'" << nurse_name.getHead() << "', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);    
  }

  // Add event to have nurse read random record (cheat)
  int cheattime = this -> time + exp_int (meanGapNurseCheat);
  addEvent (new CheatNurse(cheattime, eventCount++, this -> nurse_no, currNurses[this -> nurse_no])); 


  // Add next event to replace this nurse
  int replacetime = this -> time + exp_int (meanNurseLife);
  addEvent (new ReplaceNurse(replacetime, eventCount++, this -> nurse_no));
}



void ReplaceHcw::fire ()
{
  charbuf hcw_name(256);
  charbuf stmt(1024);

  if (currHcws[this -> hcw_no] != -1) {
    // A current hcw exists, so we should end his role
    hcw_name << "(str2prin \"_h" << currHcws[this -> hcw_no] << "\")";
    
    stmt << "INSERT INTO endrole VALUES ("
	 << "'" << hcw_name.getHead() << "',"
	 << "'hcw', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
  }
  
  // Set new current hcw
  currHcws[this -> hcw_no] = hcwCounter++;

  // New hcw name
  hcw_name.reset();
  hcw_name << "(str2prin \"_h" << currHcws[this -> hcw_no] << "\")";

  // Insert beginrole (hcw_name, hcw, u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << hcw_name.getHead() << "',"
       << "'hcw', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);
    
  // Add confidentiality agreement (but forget to do it with a small
  // probability)
  bool forgetConfAgreement = ((uniform_double() < forgetConf) ? true: false);

  if (!forgetConfAgreement) {
    stmt.reset();
    stmt << "INSERT INTO signed_conf VALUES (" 
	 << "'" << hcw_name.getHead() << "', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);    
  }

  // Add event to have hcw read random record (cheat)
  int cheattime = this -> time + exp_int (meanGapHcwCheat);
  addEvent (new CheatHcw(cheattime, eventCount++, this -> hcw_no, currHcws[this -> hcw_no])); 


  // Add next event to replace this hcw
  int replacetime = this -> time + exp_int (meanHcwLife);
  addEvent (new ReplaceHcw(replacetime, eventCount++, this -> hcw_no));
}


void ReplaceAdminStaff::fire ()
{
  charbuf adminstaff_name(256);
  charbuf stmt(1024);

  if (currAdminStaffs[this -> adminstaff_no] != -1) {
    // A current adminstaff exists, so we should end his role
    adminstaff_name << "(str2prin \"_a" << currAdminStaffs[this -> adminstaff_no] << "\")";
    
    stmt << "INSERT INTO endrole VALUES ("
	 << "'" << adminstaff_name.getHead() << "',"
	 << "'adminstaff', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
  }
  
  // Set new current adminstaff
  currAdminStaffs[this -> adminstaff_no] = adminstaffCounter++;

  // New adminstaff name
  adminstaff_name.reset();
  adminstaff_name << "(str2prin \"_a" << currAdminStaffs[this -> adminstaff_no] << "\")";

  // Insert beginrole (adminstaff_name, adminstaff, u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << adminstaff_name.getHead() << "',"
       << "'adminstaff', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);
    
  // Add confidentiality agreement (but forget to do it with a small
  // probability)
  bool forgetConfAgreement = ((uniform_double() < forgetConf) ? true: false);

  if (!forgetConfAgreement) {
    stmt.reset();
    stmt << "INSERT INTO signed_conf VALUES (" 
	 << "'" << adminstaff_name.getHead() << "', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);    
  }

  // Add event to have adminstaff read random record (cheat)
  int cheattime = this -> time + exp_int (meanGapAdminStaffCheat);
  addEvent (new CheatAdminStaff(cheattime, eventCount++, this -> adminstaff_no, currAdminStaffs[this -> adminstaff_no])); 


  // Add next event to replace this adminstaff
  int replacetime = this -> time + exp_int (meanAdminStaffLife);
  addEvent (new ReplaceAdminStaff(replacetime, eventCount++, this -> adminstaff_no));
}


void ReplaceVolunteer::fire ()
{
  charbuf volunteer_name(256);
  charbuf stmt(1024);

  if (currVolunteers[this -> volunteer_no] != -1) {
    // A current volunteer exists, so we should end his role
    volunteer_name << "(str2prin \"_v" << currVolunteers[this -> volunteer_no] << "\")";
    
    stmt << "INSERT INTO endrole VALUES ("
	 << "'" << volunteer_name.getHead() << "',"
	 << "'volunteer', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
  }
  
  // Set new current volunteer
  currVolunteers[this -> volunteer_no] = volunteerCounter++;

  // New volunteer name
  volunteer_name.reset();
  volunteer_name << "(str2prin \"_v" << currVolunteers[this -> volunteer_no] << "\")";

  // Insert beginrole (volunteer_name, volunteer, u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << volunteer_name.getHead() << "',"
       << "'volunteer', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);
    
  // Add confidentiality agreement (but forget to do it with a small
  // probability)
  bool forgetConfAgreement = ((uniform_double() < forgetConf) ? true: false);

  if (!forgetConfAgreement) {
    stmt.reset();
    stmt << "INSERT INTO signed_conf VALUES (" 
	 << "'" << volunteer_name.getHead() << "', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);    
  }

  // Add event to have volunteer read random record (cheat)
  int cheattime = this -> time + exp_int (meanGapVolunteerCheat);
  addEvent (new CheatVolunteer(cheattime, eventCount++, this -> volunteer_no, currVolunteers[this -> volunteer_no])); 


  // Add next event to replace this volunteer
  int replacetime = this -> time + exp_int (meanVolunteerLife);
  addEvent (new ReplaceVolunteer(replacetime, eventCount++, this -> volunteer_no));
}


void ReplaceDoctor::fire ()
{
  charbuf doc_name(256);
  charbuf stmt(1024);

  if (currDoctors[this -> doc_no] != -1) {
    // There was a previous doctor, so we must ends its connection
    // with all its patients
    
    int doc_id = currDoctors[this -> doc_no];
    doc_name << "(str2prin \"_d" << doc_id << "\")";

    for (int i = 0; i < nResident; i++) {
      if (doctorOf[i] == this -> doc_no) {
	// Resident i had this doctor
	
	// Insert endrole (doc_name, (doctor res_name), u) into database
	stmt.reset();
	stmt << "INSERT INTO endrole VALUES (" 
	     << "'" << doc_name.getHead() << "',"
	     << "'(doctor (str2prin \"_r" << currResidents[i] << "\"))', '" 
	     << intTimeToString (this -> time)
	     << "');";
	run_no_result(stmt);
      }
    }
  }

  // New doctor
  int doc_id = doctorCounter++;
  currDoctors[this -> doc_no] = doc_id;
  doc_name.reset();
  doc_name << "(str2prin \"_d" << doc_id << "\")";

  // Connect doctor to all patients of old doctor

  for (int i = 0; i < nResident; i++) {
    if (doctorOf[i] == this -> doc_no) {
      // Resident i had this doctor
      
      // Insert beginrole (doc_name, (doctor res_name), u) into database
      stmt.reset();
      stmt << "INSERT INTO beginrole VALUES (" 
	   << "'" << doc_name.getHead() << "',"
	   << "'(doctor (str2prin \"_r" << currResidents[i] << "\"))', '" 
	   << intTimeToString (this -> time)
	   << "');";
      run_no_result(stmt);

      // Add event to switch this doctor
      int switchtime = this -> time + exp_int (meanGapResidentSwitchDoctor);
      addEvent (new SwitchDoctor(switchtime, eventCount ++, i, currResidents[i], currDoctors[doctorOf[i]]));
    }
  }

  // Add an event to replace this doctor later
  int replacetime = this -> time + exp_int (meanDoctorLife);
  addEvent (new ReplaceDoctor(replacetime, eventCount++, this -> doc_no));
      
}

void ReplaceResident::fire () 
{
  // Statement to end the role of the current resident if it is not
  // -1. A -1 indicates the initial state, so there is no role to
  // end. Resident names begin with _r.

  charbuf res_name(256);
  charbuf doc_name(256);
  charbuf stmt(1024);
  
  int doc_no = doctorOf[this -> res_no];
  int doc_id = currDoctors[doc_no];
  
  doc_name << "(str2prin \"_d" << doc_id << "\")";

  if (currResidents[this -> res_no] != -1) {
    res_name << "(str2prin \"_r" << currResidents[this -> res_no] << "\")";
    
    // Insert endrole (res_name, resident, u) into database
    stmt << "INSERT INTO endrole VALUES (" 
	 << "'" << res_name.getHead() << "',"
	 << "'resident', '" 
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);

    if (doc_id != -1) {
      // Insert endrole (doc_name, (doctor res_name), u) into database
      stmt.reset();
      stmt << "INSERT INTO endrole VALUES ("
	   << "'" << doc_name.getHead() << "',"
	   << "'(doctor " << res_name.getHead() << ")', '"
	   << intTimeToString (this -> time)
	   << "');";
      run_no_result(stmt);
    }
  }

  // Name of new resident. 
  res_name.reset();
  res_name << "(str2prin \"_r" << residentCounter << "\")";

  // Name of personal representative of r. Pr names begin with pr.
  charbuf pr_name (256);
  pr_name << "(str2prin \"_pr" << prCounter << "\")";

  currResidents[this -> res_no] = residentCounter++;
  currPrs[this -> res_no] = prCounter++;

  // Insert beginrole (res_name, resident, u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << res_name.getHead() << "',"
       << "'resident', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);
  
  // Insert beginrole (pr_name, (pr res_name), u) into database
  stmt.reset();
  stmt << "INSERT INTO beginrole VALUES (" 
       << "'" << pr_name.getHead() << "',"
       << "'(pr " << res_name.getHead() << ")', '" 
       << intTimeToString (this -> time)
       << "');";
  run_no_result(stmt);

  if (doc_id != -1) {
    // Insert beginrole (doc_name, (doctor res_name), u) into database
    stmt.reset();
    stmt << "INSERT INTO beginrole VALUES ("
	 << "'" << doc_name.getHead() << "',"
	 << "'(doctor " << res_name.getHead() << ")', '"
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
  }
  
  // Reset the active records of resident
  activeRecords[this -> res_no] -> reset();

  // Add an event to add a new careplan for the resident. This must be
  // done at the current time.
  addEvent (new AddCareplan(this -> time, eventCount ++, res_no, currResidents[res_no]));
  
  // Add static info about patient
  int personal_id = infoCounter ++;
  int financial_id = infoCounter ++;
  int emergency_id = infoCounter ++;
  int pastmedical_id = infoCounter ++;

  stmt.reset();
  stmt << "INSERT INTO hasattr VALUES ("
       << "'(str2infoid \"_i" << personal_id << "\")',"
       << "'(rec/static/personal " << res_name.getHead() << ")'"
       << ");";
  run_no_result(stmt);

  stmt.reset();
  stmt << "INSERT INTO hasattr VALUES ("
       << "'(str2infoid \"_i" << financial_id << "\")',"
       << "'(rec/static/financial " << res_name.getHead() << ")'"
       << ");";
  run_no_result(stmt);

  stmt.reset();
  stmt << "INSERT INTO hasattr VALUES ("
       << "'(str2infoid \"_i" << emergency_id << "\")',"
       << "'(rec/static/emergency " << res_name.getHead() << ")'"
       << ");";
  run_no_result(stmt);

  stmt.reset();
  stmt << "INSERT INTO hasattr VALUES ("
       << "'(str2infoid \"_i" << pastmedical_id << "\")',"
       << "'(rec/static/pastmedical " << res_name.getHead() << ")'"
       << ");";
  run_no_result(stmt);
  
  // Add all the above static information into the system. The
  // addition is done by a randomly chosen adminstaff.
  int adminstaff_no = uniform_int(0, nAdminStaff);
  charbuf adminstaff_name (256);
  
  adminstaff_name << "(str2prin \"_a" << currAdminStaffs[adminstaff_no] << "\")";

  stmt.reset();
  stmt << "INSERT INTO does VALUES (" 
       << "'(add " << adminstaff_name.getHead() 
       << " (str2infoid \"_i" << personal_id << "\"))', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);
  
  stmt.reset();
  stmt << "INSERT INTO does VALUES (" 
       << "'(add " << adminstaff_name.getHead() 
       << " (str2infoid \"_i" << financial_id << "\"))', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);
  
  stmt.reset();
  stmt << "INSERT INTO does VALUES (" 
       << "'(add " << adminstaff_name.getHead() 
       << "(str2infoid \"_i" << emergency_id << "\"))', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);
  
  stmt.reset();
  stmt << "INSERT INTO does VALUES (" 
       << "'(add " << adminstaff_name.getHead() 
       << "(str2infoid \"_i" << pastmedical_id << "\"))', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);

  // Initially, resident is capable
  incapable[this -> res_no] = false;

  // Add event to make resident incapable later
  int incaptime = this -> time + exp_int (meanIncapabilityGap);
  addEvent (new MakeResidentIncapable(incaptime, eventCount++, this -> res_no, currResidents[this -> res_no]));

  // Add event to have resident read his record (legitimately)
  int readtime = this -> time + exp_int (meanGapResidentRead);
  addEvent (new ReadResident(readtime, eventCount++, this -> res_no, currResidents[this -> res_no]));
 
  // Add event to have resident read random record (cheat)
  int cheattime = this -> time + exp_int (meanGapResidentCheat);
  addEvent (new CheatResident(cheattime, eventCount++, this -> res_no, currResidents[this -> res_no])); 


  // Add event to switch this resident's doctor
  int switchtime = this -> time + exp_int (meanGapResidentSwitchDoctor);
  addEvent (new SwitchDoctor(switchtime, eventCount ++, this -> res_no, 
			     currResidents[this->res_no], currDoctors[doctorOf[this->res_no]]));


  // Add next event to replace this resident
  int replacetime = this -> time + exp_int (meanResidentLife);
  addEvent (new ReplaceResident(replacetime, eventCount++, this -> res_no));
}



void SwitchDoctor::fire () 
{
  if ((currResidents[this -> res_no] == this -> res_id) &&
      (currDoctors[doctorOf[this -> res_no]] == this -> doc_id)) {
    // The doctor-resident relationship still exists, so reassign

    charbuf stmt(1024);

    // Insert endrole (doc_name, (doctor res_name), u) into database
    stmt.reset();
    stmt << "INSERT INTO endrole VALUES (" 
	 << "'(str2prin \"_d" << doc_id << "\")',"
	 << "'(doctor (str2prin \"_r" << currResidents[res_no] << "\"))', '" 
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);

    // Select new doctor
    int doc_no = uniform_int (0, nDoctor);
    
    doctorOf[this -> res_no] = doc_no;

    // Insert beginrole (doc_name, (doctor res_name), u) into database
    stmt.reset();
    stmt << "INSERT INTO beginrole VALUES (" 
	 << "'(str2prin \"_d" << currDoctors[doc_no] << "\")',"
	 << "'(doctor (str2prin \"_r" << currResidents[res_no] << "\"))', '" 
	 << intTimeToString (this -> time)
	 << "');";
    run_no_result(stmt);
    
    // Add event to switch this resident's doctor again
    int switchtime = this -> time + exp_int (meanGapResidentSwitchDoctor);
    addEvent (new SwitchDoctor(switchtime, eventCount ++, this -> res_no, 
			       currResidents[this->res_no], currDoctors[doc_no]));
  }
  // Else, ignore because either doctor or resident or their
  // relationship has changed
}

void MakeResidentIncapable::fire()
{
  if (currResidents[this -> res_no] == this -> res_id) {
    // Resident has not left in meantime
    incapable[this -> res_no] = true;
    
    // Add entry into the subj_begin_incapable table
    charbuf stmt(1024);
    stmt << "INSERT INTO subj_begin_incapable VALUES ("
	 << "'(str2prin \"_r" << currResidents[this -> res_no] << "\")', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add an event to make the resident capable again
    int captime = this -> time + exp_int (meanIncapabilityLength);
    addEvent (new MakeResidentCapable(captime, eventCount++, this -> res_no, currResidents[this -> res_no]));
  }
  // else ignore the event, because resident has left
}


void MakeResidentCapable::fire()
{
  if (currResidents[this -> res_no] == this -> res_id) {
    // Resident has not left in meantime
    incapable[this -> res_no] = false;
    
    // Add entry into the subj_end_incapable table
    charbuf stmt(1024);
    stmt << "INSERT INTO subj_end_incapable VALUES ("
	 << "'(str2prin \"_r" << currResidents[this -> res_no] << "\")', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add an event to make the resident incapable again
    int incaptime = this -> time + exp_int (meanIncapabilityGap);
    addEvent (new MakeResidentIncapable(incaptime, eventCount++, this -> res_no, currResidents[this -> res_no]));
  }
  // else ignore the event, because resident has left
}

void ReadResident::fire()
{
  //   cerr << "Firing ReadResident event. Parameters are: "
  //        << "time " << time << ", priority " << priority 
  //        << ", res_no " << res_no << "\n";

  if (currResidents[this -> res_no] == this -> res_id) {
    // Resident has not left in meantime
    
    // Select a recent random record and read it

    int nRecords = activeRecords[this -> res_no] -> length();

    if (nRecords > 0) {
      int rec_no = uniform_int(0, nRecords);
      int info_id = (*activeRecords[this -> res_no])[rec_no];
      
      charbuf stmt(1024);
      stmt << "INSERT INTO does VALUES (" 
	   << "'(read (str2prin \"_r" << this -> res_id 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
    }

    // Add next event to have resident read his record (legitimately)
    int readtime = this -> time + exp_int (meanGapResidentRead);
    addEvent (new ReadResident(readtime, eventCount++, this -> res_no, currResidents[this -> res_no]));
  
  }
  // else ignore the event, because resident has left
}


void CheatResident::fire()
{
  //   cerr << "Firing CheatResident event. Parameters are: "
  //        << "time " << time << ", priority " << priority 
  //        << ", res_no " << res_no << "\n";

  if (currResidents[this -> res_no] == this -> res_id) {
    // Resident has not left in meantime

    // Insert read on a random info
    int info_id = uniform_int (0, infoCounter);
    
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_r" << this -> res_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add next event to have resident read random record (cheat)
    int cheattime = this -> time + exp_int (meanGapResidentCheat);
    addEvent (new CheatResident(cheattime, eventCount++, this -> res_no, currResidents[this -> res_no]));
  
  }
  // else ignore the event, because resident has left
}

void CheatManager::fire()
{
  if (currManager == this -> man_id) {
    // Manager has not left in meantime

    // Insert read on a random info
    int info_id = uniform_int (0, infoCounter);
    
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_m" << this -> man_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add next event to have manager read random record (cheat)
    int cheattime = this -> time + exp_int (meanGapManagerCheat);
    addEvent (new CheatManager(cheattime, eventCount++, this -> man_id));
  
  }
  // else ignore the event, because manager has left
}


void CheatNurse::fire()
{
  if (currNurses[this -> nurse_no] == this -> nurse_id) {
    // Nurse has not left in meantime

    // Insert read on a random info
    int info_id = uniform_int (0, infoCounter);
    
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_n" << this -> nurse_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add next event to have nurse read random record (cheat)
    int cheattime = this -> time + exp_int (meanGapNurseCheat);
    addEvent (new CheatNurse(cheattime, eventCount++, this -> nurse_no, currNurses[this -> nurse_no]));
  }
  // else ignore the event, because nurse has left
}


void CheatHcw::fire()
{
  if (currHcws[this -> hcw_no] == this -> hcw_id) {
    // Hcw has not left in meantime

    // Insert read on a random info
    int info_id = uniform_int (0, infoCounter);
    
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_h" << this -> hcw_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add next event to have hcw read random record (cheat)
    int cheattime = this -> time + exp_int (meanGapHcwCheat);
    addEvent (new CheatHcw(cheattime, eventCount++, this -> hcw_no, currHcws[this -> hcw_no]));
  }
  // else ignore the event, because hcw has left
}

void CheatAdminStaff::fire()
{
  if (currAdminStaffs[this -> adminstaff_no] == this -> adminstaff_id) {
    // AdminStaff has not left in meantime

    // Insert read on a random info
    int info_id = uniform_int (0, infoCounter);
    
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_a" << this -> adminstaff_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add next event to have adminstaff read random record (cheat)
    int cheattime = this -> time + exp_int (meanGapAdminStaffCheat);
    addEvent (new CheatAdminStaff(cheattime, eventCount++, this -> adminstaff_no, currAdminStaffs[this -> adminstaff_no]));
  }
  // else ignore the event, because adminstaff has left
}


void CheatVolunteer::fire()
{
  if (currVolunteers[this -> volunteer_no] == this -> volunteer_id) {
    // Volunteer has not left in meantime
    
    // Insert read on a random info
    int info_id = uniform_int (0, infoCounter);
    
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_v" << this -> volunteer_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add next event to have volunteer read random record (cheat)
    int cheattime = this -> time + exp_int (meanGapVolunteerCheat);
    addEvent (new CheatVolunteer(cheattime, eventCount++, this -> volunteer_no, currVolunteers[this -> volunteer_no]));
  }
  // else ignore the event, because volunteer has left
}



void AddCareplan::fire()
{
  if (currResidents[this -> res_no] == this -> res_id) {
    // Resident has not left in meantime

    // Create a new info
    int info_id = infoCounter ++;
    
    // Add the new info.
    charbuf stmt(1024);
    stmt << "INSERT INTO does VALUES (" 
	 << "'(add (str2prin \"_m" << currManager 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    
    // Add attribute of info.
    stmt.reset();
    stmt << "INSERT INTO hasattr VALUES ("
	 << "'(str2infoid \"_i" << info_id << "\")',"
	 << "'(rec/dynamic/careplan (str2prin \"_r" << res_id << "\") " 
	 << intTimeToString (this -> time) << ")'" << ");";
    run_no_result(stmt);
    
    // Add new info to the activeRecords
    activeRecords[this -> res_no] -> push(info_id);

    // Adjustment of careplan by manager requires consent. The consent
    // can come from patient, which we assume is always the
    // case. However, if the patient is incapable, the consent must
    // come from the pr of the patient, which we assume will happen
    // only with probability prConsents

    if (incapable[res_no] == false) {
      // Resident is capable

      stmt.reset();
      stmt << "INSERT INTO allows VALUES ("
	   << "'(str2prin \"_r" << this -> res_id << "\")'," 
	   << "'(add (str2prin \"_m" << currManager 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);

    }

    else {
      // Resident is incapable
      // We must now ask the PR of the resident

      //Does PR allow the update?
      bool allows = ((uniform_double() <= prConsents) ? true: false);
      
      if (allows) {
	stmt.reset();
	stmt << "INSERT INTO allows VALUES ("
	     << "'(str2prin \"_pr" << currPrs[res_no] << "\")'," 
	     << "'(add (str2prin \"_m" << currManager 
	     << "\") (str2infoid \"_i" << info_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);
      }
      // else we do not add the allows predicate
    }
      
      
    // Add next event to have resident read random record (cheat)
    int cpupdatetime = this -> time + exp_int (meanGapCareplanAdd);
    addEvent (new AddCareplan(cpupdatetime, eventCount++, this -> res_no, currResidents[this -> res_no]));
  }
  // else ignore the event, because resident has left
  
}


void DoctorVisit::fire ()
{
  charbuf stmt(1024);
  for (int res_no = 0; res_no < nResident; res_no ++) {
  
    if (doctorOf[res_no] != this -> doc_no) 
      // This resident is not affiliated with this doctor, so skip the
      // rest of the loop
      continue;
 

    if (uniform_double() < doctorReadsInVisit) {
      // The doctor reads a random recent record of res_no
      
      if (activeRecords[res_no] -> length() > 0) {
	int rec_no = uniform_int (0, activeRecords[res_no] -> length());
	int info_id = (*activeRecords[res_no])[rec_no];

	stmt.reset();
	stmt << "INSERT INTO does VALUES (" 
	     << "'(read (str2prin \"_d" << currDoctors[this -> doc_no] 
	     << "\") (str2infoid \"_i" << info_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);

      }
    }

    if (uniform_double() < doctorAddsInVisit) {
      // Doctor adds a new medical record

      int info_id = infoCounter ++;
      
      // Add the new info.
      stmt.reset();
      stmt << "INSERT INTO does VALUES (" 
	   << "'(add (str2prin \"_d" << currDoctors[this -> doc_no] 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
      
      // Add attribute of info.
      stmt.reset();
      stmt << "INSERT INTO hasattr VALUES ("
	   << "'(str2infoid \"_i" << info_id << "\")',"
	   << "'(rec/dynamic/medicalrecord (str2prin \"_r" << currResidents[res_no] << "\") " 
	   << "(str2prin \"_d" << currDoctors[this -> doc_no] << "\") "
	   << intTimeToString (this -> time) << ")'"
	   << ");";
      run_no_result(stmt);
    
      activeRecords[res_no] -> push(info_id);
    }

    if (uniform_double() < doctorAddsPnInVisit) {
      // Doctor adds a new private note

      int info_id = infoCounter ++;
      
      // Add the new info.
      stmt.reset();
      stmt << "INSERT INTO does VALUES (" 
	   << "'(add (str2prin \"_d" << currDoctors[this -> doc_no] 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
      
      // Add attribute of info.
      stmt.reset();
      stmt << "INSERT INTO hasattr VALUES ("
	   << "'(str2infoid \"_i" << info_id << "\")',"
	   << "'(rec/dynamic/medicalrecord/private (str2prin \"_r" 
	   << currResidents[res_no] << "\") " 
	   << "(str2prin \"_d" << currDoctors[this -> doc_no] << "\") " 
	   << intTimeToString (this -> time) << ")'"
	   << ");";
      run_no_result(stmt);
    }

    if (uniform_double() < doctorAddsPsychoInVisit) {
      // Doctor adds a new psychotherapy note

      int info_id = infoCounter ++;
      
      // Add the new info.
      stmt.reset();
      stmt << "INSERT INTO does VALUES (" 
	   << "'(add (str2prin \"_d" << currDoctors[this -> doc_no] 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
      
      // Add attribute of info.
      stmt.reset();
      stmt << "INSERT INTO hasattr VALUES ("
	   << "'(str2infoid \"_i" << info_id << "\")',"
	   << "'(rec/dynamic/medicalrecord/psycho (str2prin \"_r" 
	   << currResidents[res_no] << "\") " 
	   << "(str2prin \"_d" << currDoctors[this -> doc_no] << "\") " 
	   << intTimeToString (this -> time) << ")'" << ");";
      run_no_result(stmt);
    
      activeRecords[res_no] -> push(info_id);

      // Does the doctor allow the patient to read this psychotherapy
      // note?
      if (uniform_double() < doctorAllowsPsychoAccess) {
	stmt.reset();
	stmt << "INSERT INTO allows VALUES ("
	     << "'(str2prin \"_d" << currDoctors[this -> doc_no] << "\")'," 
	     << "'(read (str2prin \"_r" << currResidents[res_no] 
	     << "\") (str2infoid \"_i" << info_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);
      }
    }

    if (uniform_double() < doctorRecommendsPhysioInVisit) {
      int physio_id = physioCounter ++;

      // Add recommendation to trace
      stmt.reset();
      stmt << "INSERT INTO refer_physio VALUES ("
	   << "'(str2prin \"_d" << currDoctors[this -> doc_no] << "\")',"
	   << "'(str2prin \"_ph" << physio_id << "\")',"
	   << "'(str2prin \"_r" << currResidents[res_no] << "\")', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);

      addEvent(new PhysioVisit(this -> time, eventCount++, physio_id, res_no));
    }


  } //end for

  if (uniform_double() < doctorCheatsInVisit) {
    // Doctor reads a random record (cheats)

    int info_id = uniform_int (0, infoCounter);

    stmt.reset();
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_d" << currDoctors[this -> doc_no] 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);    

  }
  

  // Schedule a new visit for this doc_no
  int visittime = this -> time + periodDoctorVisit;
  addEvent (new DoctorVisit(visittime, eventCount ++, this -> doc_no));
}


void PhysioVisit::fire ()
{
  charbuf stmt(1024);

  if (uniform_double() < physioReadsInVisit) {
    // The physiotherapist reads a random recent record of res_no
      
    if (activeRecords[res_no] -> length() > 0) {
      int rec_no = uniform_int (0, activeRecords[res_no] -> length());
      int info_id = (*activeRecords[res_no])[rec_no];

      stmt.reset();
      stmt << "INSERT INTO does VALUES (" 
	   << "'(read (str2prin \"_ph" << this -> physio_id 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
    }
  }

  if (uniform_double() < physioAddsInVisit) {
    // Physiotherapist adds a new rehab record

    int info_id = infoCounter ++;
      
    // Add the new info.
    stmt.reset();
    stmt << "INSERT INTO does VALUES (" 
	 << "'(add (str2prin \"_ph" << this -> physio_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
      
    // Add attribute of info.
    stmt.reset();
    stmt << "INSERT INTO hasattr VALUES ("
	 << "'(str2infoid \"_i" << info_id << "\")',"
	 << "'(rec/dynamic/medicalrecord/rehab (str2prin \"_r" 
	 << currResidents[res_no] << "\") " 
	 << "(str2prin \"_ph" << this -> physio_id << "\") "
	 << intTimeToString (this -> time) << ")'"
	 << ");";
    run_no_result(stmt);
    
    activeRecords[res_no] -> push(info_id);
  }

  if (uniform_double() < physioAddsPnInVisit) {
    // Physiotherapist adds a new private note

    int info_id = infoCounter ++;
      
    // Add the new info.
    stmt.reset();
    stmt << "INSERT INTO does VALUES (" 
	 << "'(add (str2prin \"_ph" << this -> physio_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
      
    // Add attribute of info.
    stmt.reset();
    stmt << "INSERT INTO hasattr VALUES ("
	 << "'(str2infoid \"_i" << info_id << "\")',"
	 << "'(rec/dynamic/medicalrecord/private (str2prin \"_r" 
	 << currResidents[res_no] << "\") " 
	 << "(str2prin \"_ph" << this -> physio_id << "\") " 
	 << intTimeToString (this -> time) << ")'" << ");";
    run_no_result(stmt);
  }

  if (uniform_double() < physioCheatsInVisit) {
    // Physiotherapist reads a random record (cheats)

    int info_id = uniform_int (0, infoCounter);

    stmt.reset();
    stmt << "INSERT INTO does VALUES (" 
	 << "'(read (str2prin \"_ph" << this -> physio_id 
	 << "\") (str2infoid \"_i" << info_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);    
  }
} 



void NurseVisit::fire () 
{
  charbuf stmt(1024);

  // Id of nurse; will come in handy later
  int nurse_id = currNurses[this -> nurse_no];

  for (int res_no = 0; res_no < nResident; res_no ++) {

    if (uniform_double() < nurseReadsInVisit) {
      // The nurse reads a random recent record of res_no
      
      if (activeRecords[res_no] -> length() > 0) {
	int rec_no = uniform_int (0, activeRecords[res_no] -> length());
	int info_id = (*activeRecords[res_no])[rec_no];

	stmt.reset();
	stmt << "INSERT INTO does VALUES (" 
	     << "'(read (str2prin \"_n" << nurse_id  
	     << "\") (str2infoid \"_i" << info_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);
      }
    }

    if (uniform_double() < nurseAddsInVisit) {
      // Nurse adds a new progress record

      int info_id = infoCounter ++;
      
      // Add the new info.
      stmt.reset();
      stmt << "INSERT INTO does VALUES (" 
	   << "'(add (str2prin \"_n" << nurse_id 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
      
      // Add attribute of info.
      stmt.reset();
      stmt << "INSERT INTO hasattr VALUES ("
	   << "'(str2infoid \"_i" << info_id << "\")',"
	   << "'(rec/dynamic/progressnotes (str2prin \"_r" << currResidents[res_no] << "\") " 
	   << intTimeToString (this -> time) << ")'"
	   << ");";
      run_no_result(stmt);
    
      activeRecords[res_no] -> push(info_id);
    }
  } //end for

  // Add next nurse visit

  int visittime = this -> time + periodNurseVisit;
  addEvent (new NurseVisit(visittime, eventCount ++, this -> nurse_no));
}


void HcwVisit::fire ()
{
  charbuf stmt(1024);

  // Id of hcw; will come in handy later
  int hcw_id = currHcws[this -> hcw_no];

  for (int res_no = 0; res_no < nResident; res_no ++) {

    if (uniform_double() < hcwReadsInVisit) {
      // The hcw reads a random recent record of res_no
      
      if (activeRecords[res_no] -> length() > 0) {
	int rec_no = uniform_int (0, activeRecords[res_no] -> length());
	int info_id = (*activeRecords[res_no])[rec_no];

	stmt.reset();
	stmt << "INSERT INTO does VALUES (" 
	     << "'(read (str2prin \"_h" << hcw_id  
	     << "\") (str2infoid \"_i" << info_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);
      }
    }

    if (uniform_double() < hcwAddsInVisit) {
      // Hcw adds a new progress record

      int info_id = infoCounter ++;
      
      // Add the new info.
      stmt.reset();
      stmt << "INSERT INTO does VALUES (" 
	   << "'(add (str2prin \"_h" << hcw_id 
	   << "\") (str2infoid \"_i" << info_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
      
      // Add attribute of info.
      stmt.reset();
      stmt << "INSERT INTO hasattr VALUES ("
	   << "'(str2infoid \"_i" << info_id << "\")',"
	   << "'(rec/dynamic/progressnotes (str2prin \"_r" 
	   << currResidents[res_no] << "\") " 
	   << intTimeToString (this -> time) << ")'"
	   << ");";
      run_no_result(stmt);
    
      activeRecords[res_no] -> push(info_id);
    }
  } //end for

  // Add next hcw visit

  int visittime = this -> time + periodHcwVisit;
  addEvent (new HcwVisit(visittime, eventCount ++, this -> hcw_no));
}



void VolunteerAdd::fire ()
{
  charbuf stmt(1024);

  int volunteer_id = currVolunteers[this -> volunteer_no];

  // Pick a random resident to which progress note is added
  int res_no = uniform_int(0, nResident);

  // Create new info
  int info_id = infoCounter ++;
  
  // Add the new info
  stmt.reset();
  stmt << "INSERT INTO does VALUES (" 
       << "'(add (str2prin \"_v" << volunteer_id 
       << "\") (str2infoid \"_i" << info_id << "\"))', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);
  
  // Add attribute of info
  stmt.reset();
  stmt << "INSERT INTO hasattr VALUES ("
       << "'(str2infoid \"_i" << info_id << "\")',"
       << "'(rec/dynamic/progressnotes (str2prin \"_r" 
       << currResidents[res_no] << "\") " 
       << intTimeToString (this -> time) << ")'"
       << ");";
  run_no_result(stmt);
  
  activeRecords[res_no] -> push(info_id);

  int nexttime = this -> time + exp_int(meanGapVolunteerAdd);
  addEvent (new VolunteerAdd(nexttime, eventCount ++, this -> volunteer_no));
} 



void ExResidentRequest::fire () 
{
  charbuf stmt(1024);

  // Pick a random resident
  int res_id = uniform_int(0, residentCounter);

  // Is res_id a current resident?
  bool isActive = false;
  for (int i = 0; i < nResident; i++) {
    if (currResidents[i] == res_id) {
      isActive = true;
      break;
    }
  }

  if (!isActive) {
    // Requests can only come from ex-residents, not active ones
    
    stmt << "INSERT INTO request VALUES ("
	 << "'(str2prin \"_r" << res_id << "\")', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);

    // Schedule a response
    int responsetime = this -> time + exp_int(meanGapManagerResponse);
    addEvent(new ManagerResponse(responsetime, eventCount++, res_id));
  }

  // Schedule next request
  int reqtime = this -> time + exp_int(meanGapExResidentRequest);
  addEvent(new ExResidentRequest(reqtime, eventCount++));
}


void ManagerResponse::fire ()
{
  charbuf stmt(1024);

  int info_id = infoCounter++; // A new information. It does not
			       // matter what we send back

  stmt << "INSERT INTO does VALUES ("
       << "'(disclose (str2prin \"_m" << currManager << "\") "
       << "(str2infoid \"_i" << info_id << "\") " 
       << "(str2prin \"_r" << this -> res_id << "\"))', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);

  // Add attribute of info
  stmt.reset();
  stmt << "INSERT INTO hasattr VALUES ("
       << "'(str2infoid \"_i" << info_id << "\")',"
       << "'(rec (str2prin \"_r" << this -> res_id << "\"))');";
  run_no_result(stmt);
}



void AbuseDisclosure::fire()
{
  charbuf stmt(1024);

  // Randomly choose a resident about whom abuse is disclosed
  int res_no = uniform_int(0, nResident);

  // Check that the resident has some recent record that can be
  // disclosed
  if (activeRecords[res_no] -> length() > 0) {
    // Choose a random record
    int rec_no = uniform_int (0, activeRecords[res_no] -> length());
    int info_id = (*activeRecords[res_no])[rec_no];

    // Id of purported law official
    int law_id = lawCounter ++;
    
    // Create a disclosure record
    stmt.reset();
    stmt << "INSERT INTO does VALUES ("
	 << "'(disclose (str2prin \"_m" << currManager << "\") "
	 << "(str2infoid \"_i" << info_id << "\") "
	 << "(str2prin \"_l" << law_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);

    // Manager believes that resident is a victim of abuse. Add that
    // entry.
    stmt.reset();
    stmt << "INSERT INTO subj_believes_victim_of_abuse VALUES ("
	 << "'(str2prin \"_m" << currManager << "\")',"
	 << "'(str2prin \"_r" << currResidents[res_no] << "\")', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);

    // With probability isLawOfficial, law_id is a law official. We
    // record this.
    if (uniform_double () < isLawOfficial) {
      stmt.reset();
      stmt << "INSERT INTO subj_is_law_official VALUES ("
	   << "'(str2prin \"_l" << law_id << "\")', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);
    }    
  }

  // Add next event for abuse disclosure
  int abusetime = this -> time + exp_int (meanGapAbuseDisclosure);
  addEvent(new AbuseDisclosure(abusetime, eventCount ++));
}



///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////



void create_tables()
{

  // We create one table for each trace predicate in our system. The
  // exception is the predicate inrole, which can be defined using
  // beginrole and endrole. The character buffer cmd_buf holds our SQL
  // statements. All first-order terms are represented by their string
  // representations and, therefore, become VARCHAR2 in the
  // database. The only exceptions are time points which are
  // represented as integers wherever possible, for efficiency
  // reasons.

  // We also store in tables starting with the prefix subj_ subjective
  // predicates. This may come in handy if we start simulating a human
  // oracle during audit.

  charbuf cmd_buf(1024); // A 1K buffer should suffice.

  cmd_buf.reset();
  cmd_buf << "BEGIN TRANSACTION;";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE does (action VARCHAR2(100) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE request (principal VARCHAR2(100) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE allows (principal VARCHAR2(50) NOT NULL, action VARCHAR2(100) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE refer_physio (doctor VARCHAR2(50) NOT NULL, physio VARCHAR2(50) NOT NULL, resident VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE signed_conf (principal VARCHAR2(50) UNIQUE, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE beginrole (principal VARCHAR2(50) NOT NULL, role VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE endrole (principal VARCHAR2(50) NOT NULL, role VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";  
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE hasattr (info VARCHAR2(50) UNIQUE, attr VARCHAR2(100) NOT NULL);";  
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_begin_incapable (principal VARCHAR(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_end_incapable (principal VARCHAR(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_believes_victim_of_abuse (manager VARCHAR(50) NOT NULL, resident VARCHAR(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_is_law_official (principal VARCHAR(50) UNIQUE, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

}


// Initialize the simulation
void init_sim()
{

  create_tables();

  int i;
  for (i = 0; i < nResident; i ++) {
    currResidents[i] = -1;
    currPrs[i] = -1;
    incapable[i] = -1;
    doctorOf[i] = -1;
    activeRecords[i] = new boundedbuffer<int> (nActiveRecords);
  }

  for (i = 0; i < nDoctor; i ++) {
    currDoctors[i] = -1;
  }

  for (i = 0; i < nNurse; i++) {
    currNurses[i] = -1;
  }

  for (i = 0; i < nHcw; i++) {
    currHcws[i] = -1;
  }

  for (i = 0; i < nAdminStaff; i++) {
    currAdminStaffs[i] = -1;
  }

  for (i = 0; i < nVolunteer; i++) {
    currVolunteers[i] = -1;
  }

  currManager = -1;

  // Add manager
  addEvent(new ReplaceManager(0, eventCount++));

  // Add nurses
  for (i = 0; i < nNurse; i++) {
    addEvent(new ReplaceNurse(0, eventCount++, i));
  }

  // Add Hcws
  for (i = 0; i < nHcw; i++) {
    addEvent(new ReplaceHcw(0, eventCount++, i));
  }

  // Add AdminStaffs
  for (i = 0; i < nAdminStaff; i++) {
    addEvent(new ReplaceAdminStaff(0, eventCount++, i));
  }

  // Add Volunteers
  for (i = 0; i < nVolunteer; i++) {
    addEvent(new ReplaceVolunteer(0, eventCount++, i));
  }

  // Assign random doctors to each resident
  for (i = 0; i < nResident; i++) {
    doctorOf[i] = uniform_int(0, nDoctor);
  }

  // Add residents. Residents must be added before doctors because the
  // ReplaceDoctor event will add beginrole (doc_name, (doctor
  // res_name), u) entries. So res_name must be defined, which will be
  // done by ReplaceResident events
  for (i = 0; i < nResident; i++) {
    addEvent(new ReplaceResident(0, eventCount ++, i));
  }

  // Add Doctors
  for (i = 0; i < nDoctor; i++) {
    addEvent(new ReplaceDoctor(0, eventCount++, i));
  }

  // Start doctor visits
  for (i = 0; i < nDoctor; i++) {
    addEvent(new DoctorVisit(periodDoctorVisit, eventCount++, i));
  }
  
  // Start nurse visits
  for (i = 0; i < nNurse; i++) {
    addEvent(new NurseVisit(periodNurseVisit, eventCount++, i));
  }
  
  // Start hcw visits
  for (i = 0; i < nHcw; i++) {
    addEvent(new HcwVisit(periodHcwVisit, eventCount++, i));
  }

  // Start progressnote updates by volunteers
  for (i = 0; i < nVolunteer; i++) {
    addEvent(new VolunteerAdd(meanGapVolunteerAdd, eventCount++, i));
  }

  // Start ex-resident requests
  addEvent(new ExResidentRequest(exp_int(meanGapExResidentRequest), eventCount++));

  // Start reports to law officials about abuse victims
  addEvent(new AbuseDisclosure(exp_int(meanGapAbuseDisclosure), eventCount++));
}


// Finalize the simulation
void finalize_sim()
{
  charbuf stmt(1024);
  stmt << "END TRANSACTION;";
  run_no_result(stmt);

  for (int i = 0; i < nResident; i++) {
    delete (activeRecords[i]);
  }
}




//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////



// Initialize the database 
int init_db()
{
#ifdef USE_DB
  if (sqlite3_open_v2(dbname, &pDb, 
		      SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE,
		      NULL) != SQLITE_OK) {
    cerr << "Failed to open database: " << dbname << endl;
    return 0;
  }
#endif
  return 1;
}

// Close the database
int finalize_db()
{
#ifdef USE_DB
  if (sqlite3_close(pDb) != SQLITE_OK) {
    cerr << "Failed to close database: " << dbname << endl;
    return 0;
  }  
  pDb = NULL;
#endif
  return 1;
}


//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////


int main ()
{
  if (!init_db()) {
    finalize_db();
    return 1;
  }

  // Initialize the random number generator
  srand (1);
  
  run_sim(maxSimTime);

  if (!finalize_db()) {
    return 1;
  }

  return 0;
}
