/*
Copyright (c) <2008-2010>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef __CHARBUF_HPP
#define __CHARBUF_HPP

#include <iostream>

#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

using namespace std;

/* An abstract class of character streams. This is an interface for charstreams
 */

class charstream 
{
public:
  charstream() { }
  virtual charstream & operator << (char c) = 0;
  virtual charstream & operator << (const char * c) = 0;
  virtual charstream & operator << (int i) = 0;
  virtual ~charstream() { };
};

/* A class to hold buffers of characters. Each buffer is a bounded
   length array of characters, which can be used like a stack. The
   base end (0) is called the head, and the other the tail. There are
   functions to add and remove characters from the tail end. The tail
   end is guaranteed to be null terminated (i.e., the tail pointer
   always points to a null character), so head can be used as a null
   terminated string. The insertion functions don't keep track of
   length, and it is possible to run over the allocated length.
*/

class charbuf: public charstream
{
protected:
  char * head;
  char * tail;

public:

  // Create a buffer of length size + 1 that can hold size characters. 
  charbuf (int size) {
    head = new char[size + 1];
    head[0] = '\0';
    tail = head;
  }

  // Create a buffer by copying a string
  charbuf (const char * str) {
    int l = strlen(str);
    head = new char[l + 1];
    strcpy(head, str);
    tail = head + l;
  }

  // Add a character to the tail. 
  charbuf & operator << (char c) {
    *tail++ = c;
    *tail = '\0';
    return * this;
  }


  // Add a character string to the tail. 
  charbuf & operator << (const char * str) {
    while (*str != '\0') {
      *tail++ = *str++;
    }
    *tail = '\0';
    return * this;
  }
  
  // Add an integer to the tail. 
  charbuf & operator << (int n) {
    int pos = (n >= 0) ? n : (-n);

    char buf[20];

    buf[19] = '\0';
    int i = 18;
    while (pos > 0) {
      buf[i] = (char)(pos % 10 + '0');
      i -- ;
      pos = pos / 10;
    }

    if (n == 0) { 
      buf[i] = '0';
    }

    else if (n < 0) {
      buf [i] = '-';
    }
    
    else {
      i++;
    }

    for (; i <= 18; i++) {
      *tail++ = buf[i];
    }

    *tail = '\0';

    return * this;
  }

  // Remove from tail end, all characters upto and including the first
  // occurence of c. If c is not found, tail is set equal to head
  // (buffer is emptied). Note that if c == '\0', this call has no
  // effect.
  void rollback(char c) {
    while (tail != head && *tail != c) {
      tail --;
    }
    
    *tail = '\0';
  }

  // Use this function to convert to char *
  char * getHead() const {
    return head;
  }

  // Use this function if you want to iterate backwards.
  char * getTail() const {
    return tail;
  }


  // Use this function to reset the buffer (erase everything)
  void reset() {
    tail = head;
    head[0] = '\0';
  }
  
  int size() const {
    return tail - head;
  }

  ~charbuf() {
    delete [] head;
  }

  // Write the buffer to a file pointed by descriptor fd. Returns 1 on
  // success, 0 on failure
  int writeToFile (int fd) {
    int l = size();
    int niters = l / 512;
    int leftover = l % 512;

    char * root = head;
    for (int i = 0; i < niters; i++) {
      if (write (fd, root, 512) != 512) {
	return 0;
      }

      root = root + 512;
    }

    if (leftover > 0) {
      if (write (fd, root, leftover) != leftover) {
	return 0;
      }
    }

    return 1;
  }


};

inline ostream & operator << (ostream & out, const charbuf & cb) {
  return (out << (cb.getHead()));
}

#endif
