(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* A signature for the database that stores DB predicates. Each
   scenario must instantiate this signature. The functions to be
   provided are:

   - Opening a connection into the database for DB predicates.

   - Closing a connection into the database

   - Preparing a SQL command on an opened connection

   - Stepping a prepared command 

   - Returning the column results after each step

   - Resetting a prepared command

   - Deleting a prepared command

   We assume that the database only stores strings. These are parsed
   into structured terms. This simplifies the design of the interface
   a lot, at the cost of adding some parsing overhead.
*)

signature DB = sig

    (* Escape a string properly in the SQL syntax. For instance, in
       SQLite, the character ' in a string must be replicated because
       it is also the string delimiter. *) 
    val escape: string -> string

    (* The type of database names: could be a string for filenames or
       URIs*) 
    type dbName = string

    (* The type of database connections *)
    type dbConnection

    (* The type of compiled queries *)
    type query
	 
    (* A new connection will be created rarely *)
    val newConnection: dbName -> dbConnection Option.option
			       
    (* A dummy, do-nothing connection *)
    val dummyConnection: dbConnection 

    (* Close connection *)
    val closeConnection: dbConnection -> unit
					 
    (* Convert a string with SQL into a query *)
    val newQuery: dbConnection -> string -> query Option.option
					    
    (* Result of stepping a query: 
	     DATA -- a row is available
	     DONE -- successfully ended
	     ERROR -- error
     *)
    datatype queryresult = DATA | DONE | ERROR

    (* Step through a query *)
    val stepQuery: query -> queryresult
			    
    (* Return the row from last step on query. Null values are
       represented by ""
    *) 
    val getData: query -> string list
				   
    (* Reset a query to restart it *)
    val resetQuery: query -> unit
			     
    (* Destroy a query after it is done *)
    val destroyQuery: query -> unit
			       
end
