(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)


(* 

The abstract syntax of the declaration (signature) files. The
declarations provide static information about sorts, term constructors
and predicate symbols.

We call the sorts int, string, date primitive and others
non-primitive. Any function constructor that has a primitive sort must
have an evaluation function associated with it. The function must
return a primtive term of the sort.

TERM CONSTRUCTORS 
 - sorts of arguments
 - sort of result
 - class of constructor
   - SQL: Constructor evaluated by underlying SQL, e.g., x + y; all
     arguments and return must be primitive sorts
   - EVAL: Constructor evaluated by given function; return type must be primitive
   - SYM: Uninterpreted (symbolic) constructor; return type must be non-primitive
 - infix name (only for binary SQL constructors that must be infixed in SQL queries)

PREDICATE SYMBOLS
 - sorts of arguments
 - mode of each position: +,-,* (input, output, either)
 - the class of the symbol. One of:
   - SQL: Predicate interpreted by the underlying SQL, e.g., x = y;
     all arguments must be input mode, all arguments must be primitive
     sorts
   - EVAL: Predicate interpreted by a function (constraint)
   - DB: Database predicate (stored in the database)
   - SUBJ: Subjective predicate (not stored in the database)
 - infix name (only for binary SQL predicates that must be infixed in SQL queries)

*)


structure Declaration = struct

(* There are four pre-defined sorts: strings, integers, dates, and
any. The last one types any well-formed term
*)

datatype sort = sortString | sortInt | sortDate | sortAny | sortOther of string 

(* sortAny is for terms of arbitrary sort. There is no subsorting, so
   the signature must define explicit coercions between other sorts and
   sortAny, if they are needed 
*)
									 
datatype mode = modeInput | modeOutput 

datatype predclass = predSQL | predEval | predDB | predSubj

datatype constclass = constSQL | constSym | constEval 

(* The infix SQL name of a binary predicate or binary function *)
type sqlname = string

(* The name of a column -- needed for predDB predicates *)
type columnname = string 

(* The name of database tables -- needed for predDB predicates *)
type tablename = string 

datatype decl = declSort 
	      | declConst of sort list * sort * constclass * sqlname Option.option 
	      | declPred of (sort * mode) list * 
			    predclass * sqlname Option.option * 
			    (tablename * columnname list) Option.option

type decls = (string * decl) list

(* Is sort s primitive? *)

fun isPrimitive s = 
    case s of
	sortString => true
      | sortInt => true
      | sortDate => true
      | _ => false


(* Are sorts s1 and s2 equal? *)

fun equal_sort s1 s2 = (s1 = s2)

(* Find a declaration *)
fun lookup_decl (ds:decls) (i: string) = 
    case (List.find (fn (x, d) => x = i) ds) of
	NONE => NONE
      | SOME (_, d) => SOME d


(*************************************************************************)

(*************************************************************************)

(* Functions to convert to string *)

fun print_sort (sortString) = "string"
  | print_sort (sortInt) = "int"
  | print_sort (sortDate) = "date"
  | print_sort (sortAny) = "any"
  | print_sort (sortOther id) = id

fun print_mode modeInput = "(+)"
  | print_mode modeOutput = "(-)"

fun print_sort_list [] = ""
  | print_sort_list (so :: l) = 
    (print_sort so) ^ " -> " ^ (print_sort_list l)

fun print_sortmode_list [] = ""
  | print_sortmode_list ((so,m) :: l) = 
    (print_sort so) ^ (print_mode m) ^ " -> " ^ (print_sortmode_list l)

fun print_predclass predSQL = "SQL"
  | print_predclass predEval = "EVAL"
  | print_predclass predDB = "DB"
  | print_predclass predSubj = "SUBJ"

fun print_constclass constSQL = "SQL"
  | print_constclass constEval = "EVAL"
  | print_constclass constSym = "SYM"

fun print_infix (NONE) = ""
  | print_infix (SOME s) = " %infix " ^ "\"" ^ s ^ "\""

fun print_columnlist [] = ""
  | print_columnlist [cn] = "\"" ^ cn ^ "\""
  | print_columnlist (cn :: cnl) = 
    "\"" ^ cn ^ "\", " ^ (print_columnlist cnl)

fun print_dbinfo (NONE) = ""
  | print_dbinfo (SOME (tn, cnl)) =
    "\n%dbinfo " ^ "\"" ^ tn ^ "\" (" ^
    (print_columnlist cnl) ^ ")"

fun print_decl (id, d) = 
    case d of
	declSort => 
	id ^ ": sort."

      | declConst (sl, so, cc, inf) => 
	id ^ ": " ^ (print_sort_list sl) ^ 
	(print_sort so) ^ " " ^
	(print_constclass cc) ^ 
	(print_infix inf) ^ "."
	
      | declPred (sml, pc, inf, dbinfo) =>
	id ^ ": " ^ (print_sortmode_list sml) ^ 
	"pred " ^
	(print_predclass pc) ^ 
	(print_infix inf) ^ 
	(print_dbinfo dbinfo) ^ "."
	
fun print_decls ds = List.foldr (fn (x, s) => (print_decl x) ^ "\n" ^ s) "" ds


(*************************************************************************)

(*************************************************************************)

exception DeclNotFound of string
exception DuplicateDecl of string
exception DeclNotSort of string
exception BadInfixName of string
exception BadReturnSort of string
exception BadArgSort of string
exception BadArgMode of string
exception MissingDbInfo of string
exception BadDbInfo of string
exception NullaryDBPred of string

(* Check that a list of declarations is good *)
fun check_decls (ds: decls) =
    
    (* Has sort s been declared in past_ds? If yes, return () else raise exception *)
    let fun existsSort (past_ds: decls) (so: sort)  = 
	    case so of
		sortOther s => (* User defined sort *)
		if (List.exists (fn (s', d) => d = declSort andalso s' = s) past_ds) 
		then ()
		else (print ("Undeclared sort name: " ^ s ^ "\n");
		      raise (DeclNotSort s))
	      | _ => () (* Pre-defined sort *)
		     
	(* Helper function; checks the declarations ds, given the
           earlier declarations past_ds in reverse order. Returns ()
           on success and throws an exception otherwise *)
	
	fun checkDecls_h (ds: decls) (past_ds: decls) = 
	    case ds of 
		[] => () (* End of declarations *)
	      | (s, d) :: ds =>
		(* First, make sure that the symbol s is not already defined *)
		if (List.exists (fn (x, _) => x = s) past_ds) 
		then (print ("Duplicate declaration of identifier: " ^ s ^ "\n");
		      raise DuplicateDecl s)
		else 
		    ((case d of
			  
			  declSort => ()
				      
			| declConst(args, res, class, inName) =>
			  ((* Check that sort res is defined *)
			   (existsSort past_ds res);
			   
			   (* Check that argument sorts are defined *)
			   (List.foldr (fn (s, _) => existsSort past_ds s) () args);
			   
			   (* Check that argument sorts and return sort correspond to the class of constructor *)
      			   (case class of
				constSQL => if (isPrimitive res) 
					    then ((List.map (fn arg => if (isPrimitive arg)
								       then ()
								       else (print ("SQL constructor: " ^ s ^
										    " has argument of non-primitive sort\n");
									     raise BadArgSort s)) args);
						  ())
					    else (print ("SQL constructor: " ^ s ^ 
							 " has result of non-primitive sort\n");
						  raise BadReturnSort s
						 )
			      | constEval => if (isPrimitive res) 
					     then () 
					     else (print ("EVAL constructor: " ^ s ^ 
							  " has result of non-primitive sort\n");
						   raise BadReturnSort s
						 )

			      | constSym => if (isPrimitive res) 
					    then (print ("SYM constructor: " ^ s ^ 
							 " has result of primitive sort\n");
						  raise BadReturnSort s
						 )
					    else()
			   );
			   
			   (* Check that %infix name is given only if the constructor is SQL class and binary *)
			   (case inName of
				NONE => ()
			      | SOME _ => if ((class = constSQL) andalso (List.length args = 2)) 
					  then ()
					  else (print ("Unnecessary infix name: " ^ s ^ "\n");
						raise BadInfixName s)
			   )
			  )
			  

			| declPred (args, class, inName, dbInfo) =>
			  ((* Check that argument sorts are defined *)
			   (List.foldr (fn ((s, _), _) => existsSort past_ds s) () args);
			   
			   (* Check that if the predicate is SQL, then its arguments are primitive *)
			   (case class of
				predSQL => (List.map (fn (so,_) => 
							 if (isPrimitive so) then ()
							 else (print ("SQL predicate: " ^ s ^ 
								      " has argument of non-primitive sort\n");
							       raise BadArgSort s)) args; 
					    ()
					   )
			      | _ => ()
			   );
			   
			   (* Check that SQL predicates only have input modes *)
			   (case class of
				predSQL => (List.map (fn (_, modeInput) => ()
						       | (_, _) => (print ("SQL predicate: " ^ s ^
									   " has argument in output mode\n");
								    raise BadArgMode s)) args;
					    ()
					   )
			      | _ => ()
			   );			 

			   (* Check that %infix is given only if the predicate is SQL class and binary *)
			   (case inName of 
				NONE => ()
			      | SOME _ => if ((class = predSQL) andalso (List.length args = 2)) 
					  then ()
					  else (print ("Unnecessary infix name: " ^ s ^ "\n");
						raise BadInfixName s)
			   );
			   
			   (* Check that %dbinfo is given only iff
                              predicate is DB class, and in that case
                              verify that the length of the arguments
                              are same in the sort-mode list and the
                              dbinfo
			    *)
			   (case dbInfo of
				NONE => if (class <> predDB)
					then ()
					else (print ("%dbinfo not given for DB predicate: " ^ s ^ "\n");
					      raise MissingDbInfo s)
			      | SOME (tn, cnl) =>
				if (class <> predDB)
				then (print ("%dbinfo given for non-DB predicate: " ^ s ^ "\n");
					      raise BadDbInfo s)
				else 
				    if (List.length cnl = List.length args)
				    then ()
				    else (print ("Number of columns in %dbinfo of predicate: " ^ s ^ 
						 " differs from number of arguments\n");
					  raise BadDbInfo s)
			   );

			   (* Check that if predicate is DB, then it has at least one argument *)
			   (case class of
				predDB => if (List.length args > 0)
					  then ()
					  else (print ("Nullary DB predicate: " ^ s ^ "\n");
						raise NullaryDBPred s)
			      | _ => ()
			   )
			  )
		     );
		     (checkDecls_h ds ((s,d) :: past_ds)))
		    
    in
	checkDecls_h ds []
    end

end

