(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

%name Fol;

%tokens 
    : ID of string
    | COLON (":")
    | LPAREN ("(")
    | RPAREN (")")
    | LSQ ("[")
    | RSQ ("]")
    | LBR ("{")
    | RBR ("}")
    | LA ("<")
    | RA (">")
    | LIT_ALL ("all")
    | LIT_EX ("ex")
    | LIT_AND ("and")
    | LIT_OR ("or")
    | LIT_IMP ("imp")  (* (imp g f) is abbreviation for (all g f) -- all without any binders *)
    | LIT_TRUE ("true")
    | LIT_FALSE ("false")
    | LIT_ANY ("any")
    | LIT_DATE ("date")
    | LIT_INT ("int")
    | LIT_STRING ("string")
    | LIT_COMPILE ("COMPILE")
    | DATE of Time.time
    | STR of string
    | INT of int
    | START_FORMULA
    | START_TERM
    ;

%start st;

st
    : START_FORMULA formula => (Fol.type_formula formula)
    | START_TERM term => (Fol.type_term term)
    ;

formula
    : formula_all 
    | formula_ex
    | formula_and
    | formula_or
    | formula_true
    | formula_false
    | formula_atom
    | "(" formula ")"
    | formula_imp
    ;

guard
    : guard_atom
    | guard_and
    | guard_or
    | guard_ex
    | guard_true
    | guard_false
    | "(" guard ")"
    ;

formula_arg            (* Formula in the argument position *)
    : formula_unit
    | "(" formula ")"
    ;

formula_unit           (* A propositional letter *)
    : ID => (Fol.fAtom (Option.NONE, (ID, [])))
    | "true" => (Fol.fTrue Option.NONE)
    | "false" => (Fol.fFalse Option.NONE)
    ;


guard_arg            (* Guard in the argument position *)
    : guard_unit
    | "(" guard ")"
    ;

guard_unit           (* A propositional letter *)
    : ID => (Fol.gAtom (ID, []))
    | "true" => (Fol.gTrue)
    | "false" => (Fol.gFalse)
    ;

formula_all
    : name_opt "all" bound_list exl compile_opt guard_arg formula_arg
        => (Fol.fAll (name_opt, bound_list, exl, compile_opt, guard_arg, formula_arg))
    ;

formula_ex
    : name_opt "ex" bound_list exl compile_opt guard_arg formula_arg
        => (Fol.fEx (name_opt, bound_list, exl, compile_opt, guard_arg, formula_arg))
    ;

formula_and
    : name_opt "and" formula_arg (formula_arg)+ 
        => (Fol.fAnd (name_opt, formula_arg::SR))
    ;

formula_or
    : name_opt "or" formula_arg (formula_arg)+
        => (Fol.fOr (name_opt, formula_arg :: SR))
    ;

formula_true
    : name_opt "true" 
        => (Fol.fTrue (name_opt))
    ;

formula_false
    : name_opt "false"
        => (Fol.fFalse (name_opt))
    ;

formula_atom
    : name_opt ID term_arg_list => (Fol.fAtom(name_opt, (ID, term_arg_list)))
    ;

formula_imp
    : name_opt "imp" compile_opt guard_arg formula_arg
        => (Fol.fAll (name_opt, [], [], compile_opt, guard_arg, formula_arg))
    ; 


compile_opt (* Should this guard be compiled to SQL as a whole? *)
    : "COMPILE" => (Fol.compileYes)
    | => (Fol.compileNo)
    ;

guard_and
    : "and" guard_arg (guard_arg)+ 
        => (Fol.gAnd (guard_arg :: SR))
    ;

guard_or
    : "or" guard_arg (guard_arg)+ 
        => (Fol.gOr (guard_arg :: SR))
    ;

guard_true
    : "true" => (Fol.gTrue)
    ;

guard_false
    : "false" => (Fol.gFalse)
    ;

guard_ex
    : "ex" bound_list guard_arg
        => (Fol.gEx (bound_list, guard_arg))
    ;

guard_atom
    : ID term_arg_list => (Fol.gAtom(ID, term_arg_list))
    ;

bound_list
    : ( "[" ID (":" sort)? "]" )*
    ;
        
name_opt
    : => (Option.NONE)
    | "<" ID ">" => (Option.SOME ID)
    ;
        
sort
    : "any" => ( Declaration.sortAny )
    | "date" => ( Declaration.sortDate )
    | "int" => ( Declaration.sortInt )
    | "string" => ( Declaration.sortString )
    | ID => ( Declaration.sortOther ID )
    ;

exl
    : => ([])
    | "{" ( "[" term_arg_list  "]"  )* "}"
    ;


term
    : ID term_arg_list => (Fol.tApp (ID, term_arg_list))
    | DATE => (Fol.tDate DATE)
    | STR => (Fol.tString STR)
    | INT => (Fol.tInt INT)
    | "(" term ")"
    ;
 

term_arg_list
    : (term_arg)*
    ;

term_arg   (* Term in argument position *)
    : term_unit
    | "(" term ")"
    ;

term_unit
    : ID => (Fol.tApp (ID, []))
    | DATE => (Fol.tDate DATE)
    | STR => (Fol.tString STR)
    | INT => (Fol.tInt INT)
    ;
