(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* The syntax of terms and formulas. This is a guarded fragment of
first-order logic. *)

structure Fol = struct

(* The type of variable names *)
type var = string

(* The type of term constructor names *)
type const = string

(* The type of predicate names *)
type pred = string

(* An English description of a subformula for listing *)
type name = string 

datatype term = tVar of var (* Variable *)
	      | tString of string  (* Pre-defined terms: strings, numbers, dates *)
	      | tInt of int 
	      | tDate of Time.time
	      | tApp of const * term list


(* We segregate atoms in the syntax for quick pattern matching. The
assumption here is that the well-formedness check will transform
classify atoms into appropriate types based on the declarations *)

type atom = pred * term list

datatype guard = gAtom of atom
	       | gAnd of guard list
	       | gOr of guard list
	       | gEx of (var * Declaration.sort option) list * guard
	       | gTrue
	       | gFalse

datatype compile = compileYes | compileNo

datatype formula = fAtom of name option * atom
		 | fAnd of name option * formula list
		 | fOr of name option * formula list
		 | fEx of name option * (var * Declaration.sort option) list * term list list * compile * guard * formula 
		 | fAll of name option * (var * Declaration.sort option) list * term list list * compile * guard * formula
		 | fTrue of name option
		 | fFalse of name option

(* The term list list in quantifiers is the list of terms that are NOT
valid substitutions *)



(* Extract the name of a formula *)

fun nameOfFormula (fAtom (n, _)) = n
  | nameOfFormula (fAnd (n, _)) = n
  | nameOfFormula (fOr (n, _)) = n
  | nameOfFormula (fAll (n, _, _, _, _, _)) = n
  | nameOfFormula (fEx (n, _, _, _, _, _)) = n
  | nameOfFormula (fTrue n) = n
  | nameOfFormula (fFalse n) = n


(*************************************************************************)

(*************************************************************************)

(* Free variables of a term *)


local
    fun union [] s2 = s2
      | union (e :: s1) s2 = 
	let val s = union s1 s2
	in 
	    if (List.exists (fn x => x = e) s)
	    then s else (e :: s)
	end
in

fun fv_term t =
    case t of
	tVar v => [v]
      | tString _ => []
      | tInt _ => []
      | tDate _ => []
      | tApp (_, tl) => fv_term_list tl

and fv_term_list tl =
    List.foldr (fn (t, vl) => union (fv_term t) vl) [] tl

fun isClosed t =
    case t of
	tVar v => false
      | tString _ => true
      | tInt _ => true
      | tDate _ => true
      | tApp (_, tl) => isClosed_list tl
and isClosed_list [] = true
  | isClosed_list (t :: tl) = 
    (isClosed t) andalso (isClosed_list tl)

end

(*************************************************************************)

(*************************************************************************)

(* Functions to convert to strings *)

fun print_var v = v 

fun print_const c = c

fun print_term (tVar v) = print_var v
  | print_term (tString s) = "\"" ^ s ^ "\""
  | print_term (tInt i) = Int.toString i
  | print_term (tDate d) = Date.fmt "%Y:%m:%d:%H:%M:%S" (Date.fromTimeLocal d)
  | print_term (tApp (c, [])) = print_const c
  | print_term (tApp (c, tl)) = "(" ^ (print_const c) ^ " " ^
				(print_term_list tl) ^ ")"

and print_term_list [] = "" (* Should never be called with argument [] *)
  | print_term_list [t] = print_term t
  | print_term_list (t :: tl) = (print_term t) ^ " " ^ (print_term_list tl)


fun print_name id = "<" ^ id ^ ">"

fun print_name_opt (NONE) = ""
  | print_name_opt (SOME n) = (print_name n) ^ " "


fun print_pred p = p

fun print_atom (p, []) = print_pred p
  | print_atom (p, tl) = "(" ^ (print_pred p) ^ " " ^ (print_term_list tl) ^ ")"

fun print_bound_list [] = ""
  | print_bound_list ((id, sopt) :: bl) = 
    "[" ^ (print_var id) ^
    (case sopt of
	 NONE => ""
       | SOME so => ": " ^ (Declaration.print_sort so)) ^
    "] " ^
    (print_bound_list bl)


fun print_var_list vl = 
    print_bound_list ((List.map (fn v => (v, NONE))) vl)

fun print_guard (gAtom a) = print_atom a
  | print_guard (gAnd gl) = "(and " ^ (print_guard_list gl) ^ ")"
  | print_guard (gOr gl) = "(or " ^ (print_guard_list gl) ^ ")"
  | print_guard (gEx (vdl, g)) = "(ex " ^ (print_bound_list vdl) ^ (print_guard g) ^ ")"
  | print_guard (gTrue) = "true"
  | print_guard (gFalse) = "false"

and print_guard_list [] = ""
  | print_guard_list [g] = print_guard g
  | print_guard_list (g :: gl) = 
    (print_guard g) ^ " " ^ (print_guard_list gl)

fun print_ex_list exl =
    let fun print_ex_list_h [] = ""
	  | print_ex_list_h [tl] = "[" ^ (print_term_list tl) ^ "]"
	  | print_ex_list_h (tl :: exl) = 
	    "[" ^ (print_term_list tl) ^ "]" ^ " " ^ (print_ex_list_h exl)
    in 
	case exl of
	    [] => ""
	  | _ => "{" ^ (print_ex_list_h exl) ^ "} "
    end

fun print_compile compileYes = "COMPILE "
  | print_compile compileNo = ""

fun print_formula (fAtom (NONE, (p, []))) = print_pred p
  | print_formula (fAtom (NONE, (p, tl))) = 
    "(" ^ (print_pred p) ^ " " ^ (print_term_list tl) ^ ")"
  | print_formula (fAtom (SOME n, (p, []))) = 
    "(" ^ (print_name n) ^ " " ^ (print_pred p) ^ ")"
  | print_formula (fAtom (SOME n, (p, tl))) = 
    "(" ^ (print_name n) ^ " " ^ (print_pred p) ^ " " ^ 
    (print_term_list tl) ^ ")"

  | print_formula (fAnd (name_opt, fl)) = 
    "(" ^ (print_name_opt name_opt) ^ "and " ^
    (print_formula_list fl) ^ ")"

  | print_formula (fOr (name_opt, fl)) = 
    "(" ^ (print_name_opt name_opt) ^ "or " ^
    (print_formula_list fl) ^ ")"

  | print_formula (fEx (name_opt, vdl, exl, comp_opt, g, f)) =
    "(" ^ (print_name_opt name_opt) ^ "ex " ^ 
    (print_bound_list vdl) ^
    (print_ex_list exl) ^
    (print_compile comp_opt) ^ 
    (print_guard g) ^ " " ^
    (print_formula f) ^ ")"

  | print_formula (fAll (name_opt, vdl, exl, comp_opt, g, f)) =
    "(" ^ (print_name_opt name_opt) ^ "all " ^ 
    (print_bound_list vdl) ^
    (print_ex_list exl) ^
    (print_compile comp_opt) ^ 
    (print_guard g) ^ " " ^
    (print_formula f) ^ ")"

  | print_formula (fTrue NONE) = "true"
  | print_formula (fTrue (SOME n)) = 
    "(" ^ (print_name n) ^ " true)"
    
  | print_formula (fFalse NONE) = "false"
  | print_formula (fFalse (SOME n)) = 
    "(" ^ (print_name n) ^ " false)"
    
and print_formula_list [] = ""
  | print_formula_list [f] = print_formula f
  | print_formula_list (f :: fl) = 
    (print_formula f) ^ " " ^ (print_formula_list fl)



(*************************************************************************)

(*************************************************************************)

(* Data type that combines formulas and terms -- used as the return type of
the parser which can be used to parse both formulas and terms *)

datatype parser_types = type_formula of formula
		      | type_term of term

exception BadParseType
	  
fun toTerm (type_term t) = t
  | toTerm _ = raise BadParseType
		     
fun toFormula (type_formula f) = f
  | toFormula _ = raise BadParseType


(*************************************************************************)

(*************************************************************************)

(* Pre-processor for parsed output. Does three things:

   - Replaces bound variables, which get parsed as tApp(c, []) with
     tVar(c)
   
   - Checks that a binding occurrence of a variable does not shadow
     another variable. This should simplify processing (esp. mode
     checking) and also make display to the user easier.

   - Checks that names of sibling nodes, if any, are not the
     same. Names play no role in the logic itself, but having the same
     name may cause problems for the display system.

*)

exception DuplicateSiblingName of name
exception DuplicateBinding of var
exception VarAsConstructor of var

local
    
    (* Check that all names in name_opt_list are unique. Return () if
       this is the case and raise exception otherwise. *)

    fun unique_name_opt_list [] = ()
      | unique_name_opt_list (NONE :: nol) = 
	unique_name_opt_list nol
      | unique_name_opt_list ((SOME n) :: nol) =
	if List.exists (fn no => no = SOME n) nol
	then (print ("Name " ^ (print_name n) ^ " appears twice among siblings\n");
	      raise DuplicateSiblingName n)
	else unique_name_opt_list nol
    

    fun preprocess_term (vl: var list) (tVar v) = (tVar v)   (* This case should not arise *)
      | preprocess_term vl (tString s) = (tString s)
      | preprocess_term vl (tInt i) = (tInt i)
      | preprocess_term vl (tDate d) = (tDate d)
      | preprocess_term vl (tApp (c, tl)) = 
	if (List.exists (fn v => v = c) vl)
	then 
	    (case tl of
		 [] => tVar c
	       | _ => (print ("Error: In term " ^ (print_term (tApp(c,tl))) ^ 
			      " " ^ c ^ " is a bound variable\n"); raise (VarAsConstructor c))
	    )
	else
	    tApp (c, preprocess_term_list vl tl)
	    
    and preprocess_term_list vl tl = List.map (preprocess_term vl) tl
				     
    fun preprocess_atom vl (p, tl) = (p, preprocess_term_list vl tl)
				     
    fun preprocess_bound_list vl [] = vl
      | preprocess_bound_list vl ((v, _) :: vdl) = 
	if (List.exists (fn v' => v' = v) vl) 
	then (print ("Error: Variable " ^ (print_var v) ^ 
		     " is bound twice in the same scope\n"); 
	      raise (DuplicateBinding v))
	else preprocess_bound_list (v::vl) vdl
	     
    fun preprocess_guard vl (gAtom a) = gAtom (preprocess_atom vl a)
      | preprocess_guard vl (gAnd gl) = gAnd (preprocess_guard_list vl gl)
      | preprocess_guard vl (gOr gl) = gOr (preprocess_guard_list vl gl)
      | preprocess_guard vl (gEx (vdl, g)) = 
	let val vl' = preprocess_bound_list vl vdl
    in 
	    gEx (vdl, preprocess_guard vl' g)
	end
      | preprocess_guard vl (gTrue) = gTrue
      | preprocess_guard vl (gFalse) = gFalse
				       
    and preprocess_guard_list vl gl = List.map (preprocess_guard vl) gl
				      
    fun preprocess_ex_list vl exl = 
	List.map (preprocess_term_list vl) exl


    (* Check a formula for well-formedness; return the new formula with
       tConst (f, []) replaced with (tVar f) for variables f *)

    fun preprocess_formula vl (fAtom (n, a)) = fAtom (n, preprocess_atom vl a)
      | preprocess_formula vl (fAnd (n, fl)) = 
	(unique_name_opt_list (List.map nameOfFormula fl);
	 fAnd (n, preprocess_formula_list vl fl))
      | preprocess_formula vl (fOr (n, fl)) = 
	(unique_name_opt_list (List.map nameOfFormula fl);
	 fOr (n, preprocess_formula_list vl fl))
      | preprocess_formula vl (fEx (n, vdl, exl, comp_opt, g, f)) =
	let val vl' = preprocess_bound_list vl vdl
	in 
	    fEx (n, vdl, preprocess_ex_list vl exl, comp_opt,
		 preprocess_guard vl' g, 
	     preprocess_formula vl' f)
	end
	
      | preprocess_formula vl (fAll (n, vdl, exl, comp_opt, g, f)) =
	let val vl' = preprocess_bound_list vl vdl
	in 
	    fAll (n, vdl, preprocess_ex_list vl exl, comp_opt,
	      preprocess_guard vl' g, 
	      preprocess_formula vl' f)
	end
	
      | preprocess_formula vl (fTrue n) = fTrue n
      | preprocess_formula vl (fFalse n) = fFalse n
					   
    and preprocess_formula_list vl fl = 
	List.map (preprocess_formula vl) fl

in

fun preproc_formula f = preprocess_formula [] f

end


(* Term values. These are terms that have no further
   simplification. Technically, these are just terms without any EVAL
   or SQL constructors. We require that the output of any evaluation
   of EVAL or SQL constructors, and the terms returned by the database
   be values. Consequently, we need a function to check that a term is
   a value. Note that values cannot contain variables.

   The following function checks if term t is a value of sort s in
   declarations ds. Return () on success, raise exception on failure.
*)


exception ValueHasVar of term
exception UndeclaredConstructor of const
exception EvalConstructor of const
exception SQLConstructor of const
exception SortMismatch of term
exception TooFewTerms of Declaration.sort list
exception TooManyTerms of term list

local
    open Declaration
in

fun check_value (ds: decls) (s: sort) (t: term) =
    let fun helper s t' =
	    case t' of
		tVar v => 
		(print ("Purported value " ^ (print_term t) ^ 
			" has a variable " ^ (print_var v) ^ "\n");
		 raise ValueHasVar t
		)
	      | tString _ => 
		if s = sortString then ()
		else (print ("In purported value " ^ (print_term t) ^
			     ", expected value of sort " ^ (print_sort s) ^ 
			     ". Found: " ^ (print_term t') ^ "\n");
		      raise SortMismatch t'
		     )
	      | tInt _ => 
		if s = sortInt then ()
		else (print ("In purported value " ^ (print_term t) ^
			     ", expected value of sort " ^ (print_sort s) ^ 
			     ". Found: " ^ (print_term t') ^ "\n");
		      raise SortMismatch t'
		     )
	      | tDate _ => 
		if s = sortDate then ()
		else (print ("In purported value " ^ (print_term t) ^
			     ", expected value of sort " ^ (print_sort s) ^ 
			     ". Found: " ^ (print_term t') ^ "\n");
		      raise SortMismatch t'
		     )
	      | tApp (c, tl') =>
		(case lookup_decl ds c of
		     NONE => 
		     (print ("In purported value " ^ (print_term t) ^
			    ", constructor " ^ (print_const c) ^ 
			     " is not declared\n");
		      raise UndeclaredConstructor c
		     )
		   | SOME declSort => 
		     (print ("In purported value " ^ (print_term t) ^
			     ", constructor " ^ (print_const c) ^ 
			     " is a sort\n");
		      raise UndeclaredConstructor c
		     )
		  | SOME (declPred _) =>
		     (print ("In purported value " ^ (print_term t) ^
			     ", constructor " ^ (print_const c) ^ 
			     " is a predicate\n");
		      raise UndeclaredConstructor c
		     )
		  | SOME (declConst (_, _, constSQL, _)) =>
		     (print ("In purported value " ^ (print_term t) ^
			     ", constructor " ^ (print_const c) ^ 
			     " is marked SQL\n");
		      raise SQLConstructor c
		     )
		  | SOME (declConst (_, _, constEval, _)) =>
		     (print ("In purported value " ^ (print_term t) ^
			     ", constructor " ^ (print_const c) ^ 
			     " is marked EVAL\n");
		      raise EvalConstructor c
		     )
		  | SOME (declConst (sl', s', constSym, _)) =>
		    if (s = s') then helper_list sl' tl'
		    else (print ("In purported value " ^ (print_term t) ^
				 ", expected value of sort " ^ (print_sort s) ^ 
				 ". Found: " ^ (print_term t') ^ "\n");
			  raise SortMismatch t'
			 )
		)

	and helper_list [] [] = ()
	  | helper_list [] (t' :: tl') =
	    (print ("In purported value " ^ (print_term t) ^
		    ", too many arguments starting at " ^ 
		    (print_term_list (t' :: tl')) ^ "\n");
	     raise TooManyTerms (t' :: tl')
	    )
	  | helper_list (s' :: sl') [] =
	    (print ("In purported value " ^ (print_term t) ^
		    ", missing argument of sort " ^ 
		    (print_sort s') ^ "\n");
	     raise TooFewTerms (s' :: sl')
	    )
	  | helper_list (s' :: sl') (t' :: tl') =
	    (helper s' t'; helper_list sl' tl')

    in
	helper s t
    end
	    
	
end

(* Substitution of terms for variables. We only allow closed terms to
   be substituted for variables, and the pre-processor already ensures
   that nested bound variables have different names, so we don't have
   to rename variables or consider cases where the variable being
   substituted coincides with a bound variable.
*)

(* fun subst_term (v:term) (x:var) (t:term) = *)
(*     case t of *)
(* 	tVar y => if x = y then v else t *)
(*       | tString _ => t *)
(*       | tDate _ => t *)
(*       | tInt _ => t *)
(*       | tApp (c, tl) => tApp (c, subst_term_list v x tl) *)

(* and subst_term_list v x tl = List.map (subst_term v x) tl *)

(* fun subst_atom (v: term) (x: var) ((p, tl): atom) = *)
(*     (p, subst_term_list v x tl) *)

(* fun subst_guard v x (g: guard) = *)
(*     case g of *)
(* 	gAtom a => gAtom (subst_atom v x a) *)
(*       | gAnd gl => gAnd (subst_guard_list v x gl) *)
(*       | gOr gl => gOr (subst_guard_list v x gl) *)
(*       | gEx (vdl, g) => gEx (vdl, subst_guard v x g) *)
(*       | gTrue => gTrue *)
(*       | gFalse => gFalse *)

(* and subst_guard_list v x gl = List.map (subst_guard v x) gl *)

(* fun subst_formula v x (f: formula) =  *)
(*     case f of *)
(* 	fAtom (nopt, a) => fAtom (nopt, subst_atom v x a) *)
(*       | fAnd (nopt, fl) => fAnd (nopt, subst_formula_list v x fl) *)
(*       | fOr (nopt, fl) => fOr (nopt, subst_formula_list v x fl) *)
(*       | fEx (nopt, vdl, exll, copt, g, f) => *)
(* 	fEx (nopt, vdl, List.map (subst_term_list v x) exll, copt, *)
(* 	     subst_guard v x g, subst_formula v x f) *)
(*       | fAll (nopt, vdl, exll, copt, g, f) => *)
(* 	fAll (nopt, vdl, List.map (subst_term_list v x) exll, copt, *)
(* 	      subst_guard v x g, subst_formula v x f) *)
(*       | fTrue nopt => fTrue nopt *)
(*       | fFalse nopt => fFalse nopt *)
	     
(* and subst_formula_list v x fl = List.map (subst_formula v x) fl *)

end
