(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
Copyright (c) <2011>, <Limin Jia, liminjia@cmu.edu>

All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*************************************************************************)

(*************************************************************************)

(* Implementation of the SCENARIO signature for the Bogeholz case
   study. Mathematically, this file implements a structure for the
   signature in the file bogeholz.d.
*)

structure HipaaScenario :> SCENARIO = 
struct

structure Db = SQLiteDb



val db_name: Db.dbName = "hipaa.db"

val declarations: Declaration.decls = 
    let val declFileName =  "./hipaa-case-study/hipaa.d" 
	val ds = Option.valOf (DeclParser.parse_decls_file declFileName)
	val _ = Declaration.check_decls ds
    in
	ds
    end


local
    open Fol
    open Time

    exception BadInput
    exception CannotOpenConnection
    exception NotPurpose
    exception Impossible

    (* Convert n days to Time.time *)
    fun fromDays (n: int): Time.time = 
	let val secs = (LargeInt.fromInt n) * (LargeInt.fromInt 86400)
	in
	    Time.fromSeconds secs
	end
	
    (* Not required in the signature, but we need it to implement hasattr
       and inrole predicates *) 
    structure Sql = SQLFn (Db)

in
    fun eval_term c tl = 
	case (c, tl) of
	    ("t_reasonable", _) => tInt (30) 
	  | ("prompt", _) => tInt (5)
	  | ("plus", [tDate (t: Time.time), tInt n]) =>	tDate (t + (fromDays n))
	  | _ => (print ("In HipaaScenario.eval_term: Term " ^ 
			 (print_term (tApp(c,tl))) ^ 
			 " not understood\n");
		  raise BadInput)

   (* Return the super purposes of a given purpose *)
    fun super_purps (t: term) = 
	case t of
	    tApp ("treatment", [k]) => [tApp ("Treatment", [])]
	  | tApp ("payment", [k]) => [tApp ("Payment", [])]
	  | tApp ("healthcare-op", [k]) => [tApp ("Healthcare-op", [])]
	  | tApp ("healthcare-op-para-1-2", [k]) => [tApp ("Healthcare-op", [])]
	  | tApp ("law-enforcement-relevant-id-or-location", [k]) => [tApp ("law-enforcement", [])]
	  | tApp ("suspicious-death-notification", [k]) => [tApp ("law-enforcement", [])]
	  | tApp ("report-possible-crime-on-premises", [k]) => [tApp ("law-enforcement", [])]
	  | tApp ("alert-of-crime-commission-location-victims-perpetrator", [k]) => [tApp ("law-enforcement", [])]
	  | _ => []

   (* Return the super roles of a given role *)
    fun super_roles (t: term) = 
	case t of
	    tApp ("health-care-provider", [k]) => [tApp ("Health-care-provider", [])]
	  | _ => []
 
    fun resolve_eval_pred connection p tl sc = 
	case (p, tl) of
	    ("inrole", [k, r, tDate T]) =>
	    (* k, r, u are input mode here *)
	    let 
		    
		val _ = 
		    let fun call_all sc [] = raise Impossible
			  | call_all sc [role] = sc (Solution.NEXT [k, role, tDate T])
			  | call_all sc (role :: rl) =
			    (sc (Solution.NEXT [k, role, tDate T]); call_all sc rl)
		    in Sql.resolve_db_pred declarations connection "inrole_db" [k, tVar "_", tDate T]
					   (fn Solution.NOMORE =>  sc Solution.NOMORE
					     | Solution.NEXT [_, role ,_] =>  
					       let val rl = super_roles role
					       in
						   call_all sc (role :: rl)
					       end
					     | Solution.NEXT tl => 
					       (print ("Unexpected output from inrole_db table: " ^
						       (print_atom ("inrole_db", tl)) ^ "\n");
						raise Impossible)
					   )
		    end


	    (* We need to check whether k is in role r at time T or
	       not. Our strategy is: 

               1. Find the set U_s = {u_s | beginrole (k, r, u_s)}
               2. Find the largest u_s \in U_s such that 0 <= u_s <= T
                  (no such u_s => inrole(k,r,u) does NOT hold)
               3. Find the set U_e = {u_e | endrole (k, r, u_e)}
               4. Find a u_e \in U_e such that u_s < u_e <= T
                  (such u_e exists => inrole(k,r,u) does NOT hold, else it does)
	     *)

		val u_s = ref (Time.zeroTime) 
		(* u_s will eventually contain the largest u satisfying beginrole(k,r,u) and 0 <= u <= T *) 
			  
		val exists_s = ref (false)
		(* Will be true iff we find such a u_s *)
			       
		val _ = Sql.resolve_db_pred declarations connection "beginrole" [k, r, tVar "_"]
					    (fn Solution.NOMORE => ()
					      | Solution.NEXT [_,_,tDate u] => 
						if (u >= !u_s andalso u <= T) then (u_s := u; exists_s := true) else ()
					      | Solution.NEXT tl => raise Impossible
					    )
	    in
		if (!exists_s = false) 
		then
		    (* There wasn't a beginrole (k,r,...) event before T *)
		    sc (Solution.NOMORE)
		else
                    (* !u_s is the largest u satisfying beginrole(k,r,u) and 0 <= u <= T *)
		    let 			 
			local
			    exception Found
			in
			
			(* exists_e is true iff there is a u_e such that endrole(k,r,u_e) and T >= u_e > u_s *)
			val exists_e = 
			    (Sql.resolve_db_pred declarations connection "endrole" [k, r, tVar "_"]
						 (fn Solution.NOMORE => ()
						   | Solution.NEXT [_,_,tDate u] => 
						     if (u <= T andalso u > !u_s) then raise Found else ()
						   | Solution.NEXT tl => raise Impossible
						 );
			     false) handle Found => true
			end
			
		    in
			if (exists_e) 
			then (* The role ended before or at T *)
			    sc (Solution.NOMORE)
			else
			    (sc (Solution.NEXT [k, r, tDate T]);
			     sc (Solution.NOMORE))
		    end
	    end 
	    
	  | ("hasattrof", [i,_ , _]) => 
	    (* t1 is input mode here *)
	    Sql.resolve_db_pred declarations connection "hasattrof_db" [i, tVar "_" , tVar "_", tVar "_"]
		            (fn Solution.NOMORE => sc (Solution.NOMORE)
			      | Solution.NEXT [t1, t2, t3, t4] =>  
				(sc (Solution.NEXT [t1, t2, t3]);
				 sc (Solution.NOMORE))
			      | Solution.NEXT tl =>  	
				(print ("Unexpected output from hasattrof_db table: " ^
					(print_atom ("hasattrof_db", tl)) ^ "\n");
				 raise Impossible
				)
			    )
	  | ("sub_attr", [ _, t1]) => (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))
	  (* this is mostly fake for now *)
					  
	  | ("attr_in", [ t1, t2]) => 
	    Sql.resolve_db_pred declarations connection "attr_in_db" [t1, t2, tVar "_"]
		            (fn Solution.NOMORE => sc (Solution.NOMORE)
			      | Solution.NEXT [t1, t2, _] =>  
				(sc (Solution.NEXT [t1, t2]);
				 sc (Solution.NOMORE))
			      | Solution.NEXT tl =>  	
				(print ("Unexpected output from attr_in_db table: " ^
					(print_atom ("attr_in_db", tl)) ^ "\n");
				 raise Impossible
				)
			    ) (**can be extended to deal with attr hierachy if needed*)

	  | ("purp_in", [t1, t2]) => 
	    let 
		(* Call the success continuation with all solutions in list *)
		fun call_all sc [] = raise Impossible
		  | call_all sc [purp] = sc (Solution.NEXT [t1,purp])
		  | call_all sc (purp :: pl) =
		    (sc (Solution.NEXT [t1, purp]); call_all sc pl)
	    in
	    Sql.resolve_db_pred declarations connection "purp_in_db" [t1, tVar "_", tVar "_"]
		            (fn Solution.NOMORE => sc (Solution.NOMORE)
			      | Solution.NEXT [_, purp, _] =>  
				let val pl = super_purps purp
				in
				    call_all sc (purp :: pl)
				end
			      | Solution.NEXT tl =>  	
				(print ("Unexpected output from subj_purp_in table: " ^
					(print_atom ("purp_in", tl)) ^ "\n");
				 raise Impossible
				)
			    )
	    end 

	  | ("eq_prin", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("eq_msg", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("eq_time", [t1, _]) => 
	    (* t1 is input mode here *)
	    (sc (Solution.NEXT [t1,t1]); sc (Solution.NOMORE))

	  | ("lessthan", [tDate t1, tDate t2]) => 
            (* t1 and t2 are both in input mode here *)
	    if (t1 <= t2) 
	    then
		(sc (Solution.NEXT [tDate t1, tDate t2]); 
		 sc (Solution.NOMORE))
	    else 
		sc (Solution.NOMORE)
	  | ("disclose-eval", [p1, p2, m, t]) => 
	    Sql.resolve_db_pred declarations connection "disclose" [p1, p2, m, t]
		            (fn Solution.NOMORE => sc (Solution.NOMORE)
			      | Solution.NEXT [q1, q2, mm, tt] =>  
				 (sc (Solution.NEXT [q1, q2, mm, tt ]);
				  sc (Solution.NOMORE))
			      | Solution.NEXT tl =>  	
				(print ("Unexpected output from hasattrof_db table: " ^
					(print_atom ("disclose", tl)) ^ "\n");
				 raise Impossible
				)
			    )

	  | ("time_in", [tDate t0, tDate t1, tDate t2]) => 
            (* t0, t1 and t2 are all in input mode here *)
	    if (t0 < t1) andalso (t1 <= t2) 
	    then
		(sc (Solution.NEXT [tDate t0, tDate t1, tDate t2]); 
		 sc (Solution.NOMORE))
	    else 
		sc (Solution.NOMORE)

	  | _ => (print ("In HipaaScenario.resolve_eval_pred: Atom " ^ 
			 (print_atom (p,tl)) ^ 
			 " not understood\n");
		  raise BadInput)
 
val subj_pred_list_true = ref []
val subj_pred_list_false = ref []

fun isFinal p tl = 
    (* Ideally, this should check the time argument of tl and compare
       to clock time; here we will assume that databases are complete,
       so everything is always final *)
    if p = "disclose-eval" then false
    else true

end (* local *)

end
