(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* Mode checking for FOL. In addition to standard I/O mode checking,
the procedure also checks that:

- SUBJ predicates do not appear in guards.

- For guards that are marked COMPILE (i.e., compile to SQL as a
  whole), the only allowed predicates are of classes SQL and DB.

- In arguments in output positions, variables under constructors
  marked SQL or EVAL are ground.

- For guards that are marked COMPILE, (i.e., compile to SQL as a
  whole), any variable nested under a EVAL or SYM constructor must be
  grounded *OUTSIDE* the guard taken as a whole (i.e., in a scope that
  is resolved before compilation to SQL starts).
*)

signature MC = sig
    val mc_formula: Declaration.decls -> Fol.formula -> unit
end

structure Mc: MC = struct

exception ModeCheckInternal
exception SubjInGuard of Fol.pred
exception EvalInCompileGuard of Fol.pred
exception CompileUngroundUnderNonSQL of Fol.var
exception InputNotGround of Fol.term
exception OutputTermBad of Fol.term
exception NonGroundingGuard of Fol.guard
exception OutputUngroundUnderEval of Fol.var

local
    open Declaration
    open Fol

    (* Set operations on lists *)

    fun union [] s2 = s2
      | union (e :: s1) s2 = 
	let val s = union s1 s2
	in 
	    if (List.exists (fn x => x = e) s)
	    then s else (e :: s)
	end
	   
    fun intersect [] s2 = []
      | intersect (e :: s1) s2 =
	if (List.exists (fn x => x = e) s2) 
	then e :: (intersect s1 s2) 
	else intersect s1 s2

    (* s1 \ s2 *)
    fun subtract [] s2 = []
      | subtract (e :: s1) s2 =
	if (List.exists (fn x => x = e) s2) 
	then subtract s1 s2
	else e :: (subtract s1 s2)
	
    (* Is s1 a subset of s2? *)
    fun subset s1 s2 = ((subtract s1 s2) = [])

    (* Check that all variables in t nested under non-SQL constructors
       are contained in the list vl. This helper function is needed to
       check guards that are marked COMPILE. Return () on success,
       raise exception on failure *)

    fun check_term_closed_under_non_sql ds vl t =
	(* helper function. First argument is true iff the second
           argument is nested under an outer constructor of type EVAL
           or SYM.
         *)
	let fun helper true (tVar v) = 
		if (List.exists (fn v' => v = v') vl) 
		then ()
		else (print ("Variable " ^ (print_var v) ^
			     " in term " ^ (print_term t) ^
			     " appears in COMPILE guard under non-SQL constructor," ^
			     " but is not ground outside the guard\n");
		      raise CompileUngroundUnderNonSQL v)
	      | helper false (tVar v) = ()
	      | helper _ (tString _) = ()
	      | helper _ (tInt _) = ()
	      | helper _ (tDate _) = ()
	      | helper b (tApp (c,tl)) =
		(case lookup_decl ds c of
		     SOME (declConst (_, _, constSQL, _)) =>
		     (List.map (helper b) tl; ())
		   | SOME (declConst (_, _, _, _)) =>
		     (* c is non-SQL, so move Boolean to true *)
		     (List.map (helper true) tl; ())
		   | _ => (* Should not happen *)
		     (print ("Undeclared/non-constructor: " ^ (print_const c) ^
			     " used as constructor\n");
		      raise ModeCheckInternal
		     )
		)
	in
	    helper false t
	end
	

    (* Check that variables of t under EVAL and SQL constructors are
       contained in vl. Return () on success, raise exception on
       failure.
    *)	
    fun check_output_term ds vl t =
	(* helper function. First argument is true iff the second
           argument is nested under an outer constructor of type EVAL
           or SQL.
         *)
	let fun helper true (tVar v) = 
		if (List.exists (fn v' => v = v') vl) 
		then ()
		else (print ("Variable " ^ (print_var v) ^
			     " in ouput term " ^ (print_term t) ^
			     " appears under SQL or EVAL constructor," ^
			     " but is not ground\n");
		      raise OutputUngroundUnderEval v)
	      | helper false (tVar v) = ()
	      | helper _ (tString _) = ()
	      | helper _ (tInt _) = ()
	      | helper _ (tDate _) = ()
	      | helper b (tApp (c,tl)) =
		(case lookup_decl ds c of
		     SOME (declConst (_, _, constSym, _)) =>
		     (List.map (helper b) tl; ())
		   | SOME (declConst (_, _, _, _)) =>
		     (* c is SQL or EVAL, so move Boolean to true *)
		     (List.map (helper true) tl; ())
		   | _ => (* Should not happen *)
		     (print ("Undeclared/non-constructor: " ^ (print_const c) ^
			     " used as constructor\n");
		      raise ModeCheckInternal
		     )
		)
	in
	    helper false t
	end

    (* Mode check for a term list that is the argument to a single
       predicate. vl is the list of variables that are already ground,
       vcurrl is the list of variables grounded in earlier terms in
       the current predicate (initially []), ml is the list of modes
       and tl is the list of terms to check. sc is the success
       continuation to be called with an argument that equals vl
       appended to all variables in output positions. The check to be
       made is this: if a term is in input mode, then it must be
       ground in vl. If a term is in output mode, then all variables
       in it are either in vl or appear only under SYM function
       symbols. Note that because this function is checking the term
       list of a single predicate, variables grounded earlier in the
       term list are not added to vl for checking future terms.
    *)
    fun mc_term_list (ds: decls) (vl: var list) (vcurrl: var list) 
		     (ml: mode list) (tl: term list) 
		     (sc: var list -> 'a): 'a =
	case (ml, tl) of
	    ([], []) => sc (vcurrl @ vl)
	  | (modeInput :: ml, t :: tl) =>
	    if (subset (fv_term t) vl) 
	    then mc_term_list ds vl vcurrl ml tl sc
	    else (print ("Term " ^ (print_term t) ^ 
			 " in input mode is not ground\n");
		  raise InputNotGround t)
	  | (modeOutput :: ml, t :: tl) =>
	    (check_output_term ds vl t;
	     mc_term_list ds vl (union (subtract (fv_term t) vl) vcurrl) ml tl sc)
	  | _ => 
	    (* tl and ml are different lengths; should not happen for well-typed formulas *)
	    raise ModeCheckInternal

    (* Mode check for guard atoms. ds is the signature, bv is the list
       of variables bound in the current quantifier, vl is the list of
       variables ground earlier, (p,tl) is the guard to mode check,
       and copt specifies whether or not this atom will be
       pre-compiled to SQL. sc is the success continuation to call if
       mode check on g succeeds.
    *)

    fun mc_gatom (ds: decls) (bv: var list) (vl: var list) ((p, tl): atom)
		 (copt: compile) (sc: var list -> 'a): 'a =
	(* Find the declaration for p -- it should exist and it should be
           a predicate, else the type-check would have failed. *)
	case lookup_decl ds p of
	    
	    SOME (declPred (sml, class, _, _)) =>
	    
	    ((* Check that p is not subjective, and if copt = compileYes, then 
	        p is not EVAL either 
	      *)
	     (case (class, copt) of
		  (predSubj, _) => 
		  (print ("SUBJ predicate: " ^ (print_pred p) ^ 
			  " occurs in guard\n");
		   raise SubjInGuard p)
		| (predEval, compileYes) => 
		  (print ("EVAL predicate: " ^ (print_pred p) ^
			  " occurs in guard marked COMPILE\n");
		   raise EvalInCompileGuard p)
		| _ => ()
	     );
	     
	     (* If copt = compileYes, then check that variables nested
                under SYM and EVAL constructors all appear in (vl \
                bv). The latter is the list of variables that were
                already ground in outer scopes.
              *)
	     (case copt of
		  compileNo => ()
		| compileYes => 
		  (List.map 
		       (check_term_closed_under_non_sql ds (subtract vl bv))
		       tl;
		   ()
		  )
	     );
	     
	     (* Usual mode check *)
	     (let val modelist = List.map #2 sml 
                  (* The type check ensures that modelist and tl have
		     the same length *) 
	      in 
		  mc_term_list ds vl [] modelist tl sc
	      end
	     )
	    )
	    
	  | _ => (* Should not happen due to type-checking *)
	    (print ("Mode check: Undeclared/non-predicate: " ^ p ^ 
		    " used as predicate\n");
	     raise ModeCheckInternal) 
	    
	    

    (* Mode check for guards. ds is the signature, bv is the list of
       variables in the immediate outside quantifier of the main
       formula or existential quantifiers in the guard, vl is the list
       of variables ground earlier, g is the guard to mode check, and
       copt says whether the guard is to be pre-compiled to SQL or
       not. sc is the success continuation to call if mode check on g
       succeeds. The argument of sc is vl appended to the list of
       variables ground by g. Raise exception on error.
    *)

    fun mc_guard (ds:decls) (bv: var list) (vl: var list) (g: guard)
		 (copt: compile) (sc: var list -> 'a): 'a =
	case g of 
	    gAtom a => mc_gatom ds bv vl a copt sc
	  | gAnd gl => mc_guard_list_conj ds bv vl gl copt sc
	  | gOr gl => mc_guard_list_disj ds bv vl gl copt sc
	  | gEx (vdl, g) =>
	    let val bvars = List.map #1 vdl 
	    in
		mc_guard ds (bv @ bvars) vl g copt 
			 (fn vl' => sc (subtract vl' bvars))
	    end
	  | gTrue => sc vl
	  | gFalse => sc vl
		      
    and mc_guard_list_conj ds bv vl [] copt sc = sc vl
      | mc_guard_list_conj ds bv vl (g :: gl) copt sc = 
	mc_guard ds bv vl g copt (fn vl' => 
				     mc_guard_list_conj ds bv vl' gl copt sc)
	
    and mc_guard_list_disj ds bv vl [] copt sc = sc (union bv vl)
      | mc_guard_list_disj ds bv vl (g :: gl) copt sc =
	mc_guard ds bv vl g copt (fn vl' => 
				     mc_guard_list_disj ds bv vl gl copt
							 (fn vl'' => sc (intersect vl' vl'')))
	


    (* Mode check helper function for formulas. ds is the signature,
       vl is the list of variables ground (or bound) in outer scopes,
       f is the formula to mode check. sc is the success continuation
       to call if mode check on f succeeds. Raise exception on error.
    *)

    fun mc_formula_h (ds: decls) (vl: var list) (f: formula) 
		     (sc: unit -> 'a): 'a = 

	case f of

	    fAtom (_, (_,tl)) => sc()
	    (* We had to check here that fv(tl) are in vl; but that is
               already ensured by type-checking, so this should not
               ever fail *)

	  | fAnd (_, fl) => mc_formula_list ds vl fl sc
	  | fOr (_, fl) => mc_formula_list ds vl fl sc
	  | fEx (_, vdl, _, copt, g, f) =>
	    let val bvars = List.map #1 vdl
	    in
		mc_guard ds bvars vl g copt
		(fn vl' =>
		    (* vl' is the list of variables ground after the
                       guard evaluates. Check that it includes bvars.
                    *)
		    let val diff = subtract bvars vl'
		    in
			if diff = [] 
			then mc_formula_h ds (bvars @ vl) f sc
			else (print ("Bound variables: " ^ (print_var_list diff) ^
				     " are not grounded by the guard " ^ 
				     (print_guard g) ^ "\n");
			      raise NonGroundingGuard g)
		    end
		)
	    end
	  | fAll (_, vdl, _, copt, g, f) => 
	    let val bvars = List.map #1 vdl
	    in
		mc_guard ds bvars vl g copt
		(fn vl' =>
		    (* vl' is the list of variables ground after the
                       guard evaluates. Check that it includes bvars.
                    *)
		    let val diff = subtract bvars vl'
		    in
			if diff = [] 
			then mc_formula_h ds (bvars @ vl) f sc
			else (print ("Bound variables: " ^ (print_var_list diff) ^
				     " are not grounded by the guard " ^ 
				     (print_guard g) ^ "\n");
			      raise NonGroundingGuard g)
		    end
		)
	    end
	  | fTrue _ => sc ()
	  | fFalse _ => sc ()

    and mc_formula_list ds vl [] sc = sc ()
      | mc_formula_list ds vl (f :: fl) sc =
	mc_formula_h ds vl f (fn () => mc_formula_list ds vl fl sc)



in

fun mc_formula (ds: decls) (f: formula) = 
    mc_formula_h ds [] f (fn () => ())

end	

end


