(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* A functor that defines the reduction algorithm. The input to the
  functor is a structure satisfying the signature SCENARIO, thus
  providing methods to access the database and simplify evaluable
  constructors and predicates.
*)


functor ReductionFn (Scenario: SCENARIO) :> sig
    val resolve_guard: Scenario.Db.dbConnection -> Fol.guard -> bool ref 
		       -> (Solution.subst -> unit) 
		       -> unit

    (* Perform the maximum possible reduction *)
    val reduce_formula: Scenario.Db.dbConnection -> Fol.formula -> Fol.formula

    (* Reduce, but do not instantiate quantifiers *)
    val reduce_formula_no_inst: Scenario.Db.dbConnection -> Fol.formula -> Fol.formula

    (* Reduce, but contract only trues and falses *)
    val reduce_simplify: Fol.formula -> Fol.formula

    (* If the top level connective is a quantifier, instantiate it, else do nothing *)
    val instantiate_formula: Scenario.Db.dbConnection -> Fol.formula -> Fol.formula			  
end = struct

local
    open Fol
    open Declaration

    structure Db = Scenario.Db

    (* Create a SQL structure to get access to DB predicates in the database *)
    structure Sql = SQLFn (Db)

    val declarations = Scenario.declarations

    exception Impossible
    exception Unimplemented
in


(* Apply a substitution su to a term t and recursively simplify the
   result. 
*)

fun subst_simplify_term (su: Solution.subst) (t: term) = 
    case t of
	tVar x =>
	(case Solution.find_subst su x of
	     NONE => tVar x
	   | SOME v => v
	)
      |	tString _ => t
      | tInt _ => t
      | tDate _ => t
      | tApp (c, tl) =>
	let val vl = subst_simplify_term_list su tl
	in
	    case lookup_decl declarations c of
		SOME (declConst (_, _, constSym, _)) => tApp(c, vl) (* c is SYM  *)
	      | SOME (declConst (_, _, _, _)) => Scenario.eval_term c vl (* c is EVAL/SQL *)
		(* Starting from a mode-checked formula, the terms in
                   vl cannot contain variables, so it is OK to
                   simplify here.
                 *)
	      | _ => (* Should not happen *)
		(print ("In ReductionFn.subst_simplify, purported constructor " ^
			(print_const c) ^ " has non-constructor declaration\n");
		 raise Impossible)
	end

and subst_simplify_term_list su tl = List.map (subst_simplify_term su) tl


(* Simplify a possibly open term to the extent possible *)

fun simplify_term (t: term) = subst_simplify_term []	     

(* Functions to apply a substitution to a syntactic class and then
   simplify resulting terms in it. Because the type-checker eliminates
   variable shadowing, we don't have to worry about it. *)

fun subst_simplify_atom su ((p, tl): atom) = 
    (p, subst_simplify_term_list su tl)

fun subst_simplify_guard su (g: guard) = 
    case g of
	gAtom a => gAtom (subst_simplify_atom su a)
      | gAnd gl => gAnd (subst_simplify_guard_list su gl)
      | gOr gl => gOr (subst_simplify_guard_list su gl)
      | gEx (vdl, g) => gEx (vdl, subst_simplify_guard su g)
      | gTrue => gTrue
      | gFalse => gFalse

and subst_simplify_guard_list su gl = List.map (subst_simplify_guard su) gl

fun subst_simplify_formula su (f: formula) = 
    case f of
	fAtom (nopt, a) => fAtom (nopt, subst_simplify_atom su a)
      | fAnd (nopt, fl) => fAnd (nopt, subst_simplify_formula_list su fl)
      | fOr (nopt, fl) => fOr (nopt, subst_simplify_formula_list su fl)
      | fEx (nopt, vdl, exll, copt, g, f) =>
	fEx (nopt, vdl, List.map (subst_simplify_term_list su) exll, copt,
	     subst_simplify_guard su g, Solution.subst_formula su f)
	(* We do not simplify in a quantified formula. That
           simplification will happen when the quantifier is
           instantiated *)
      | fAll (nopt, vdl, exll, copt, g, f) =>
	fAll (nopt, vdl, List.map (subst_simplify_term_list su) exll, copt,
	      subst_simplify_guard su g, Solution.subst_formula su f)
      | fTrue nopt => fTrue nopt
      | fFalse nopt => fFalse nopt

and subst_simplify_formula_list su fl = List.map (subst_simplify_formula su) fl



(* Resolve a non-SUBJ atom (p,tl) by finding all its satisfying
   substitutions. For each such substitution su, the success
   continuation is called once with (SOME su), and then finally it is
   called once with NONE.
   
   Note that this function has no input substitution. Any substitution
   from an outer context must be applied to tl before it is passed
   into the function.
*)

fun resolve_atom (connection: Db.dbConnection) (p, tl)
		 (sc: Solution.subst option -> 'a): 'a =

    case lookup_decl declarations p of
	SOME (declPred (_, predDB, _, _)) =>
	(* DB predicate *)
	Sql.resolve_db_pred declarations connection p tl
			    (fn Solution.NOMORE => sc NONE
			      | Solution.NEXT tl' =>
				sc (Solution.match_term_lists [] tl tl')
			    )
				
      | SOME (declPred (_, predSubj, _, _)) =>
	(* Subjective predicate; should not happen by mode checking *)
	(print ("In function ReductionFn.resolve_atom: Encountered subjective atom " ^
		(print_atom (p,tl)) ^ "\n");
	 raise Impossible
	)
	
      | SOME (declPred (_, _, _, _)) =>
	(* SQL/EVAL predicate *)
        Scenario.resolve_eval_pred connection p tl 	  
				   (fn Solution.NOMORE => sc NONE
				     | Solution.NEXT tl' =>
				       sc (Solution.match_term_lists [] tl tl')
				   )
      | _ => 
	(print ("In function ReductionFn.resolve_atom: Purported predicate " ^
		(print_pred p) ^ "is not a declared predicate\n");
	 raise Impossible
	)


(* Resolve a guard g by finding all its satisfying substitutions. su
   is the incoming substitution. For each satisfying substitution su'
   that extends su, the success continuation sc is called once with
   su'. Set the ref cell isFinal to false if any one of the resolution
   of atoms might result in more substitutions in future extensions.

   The idea here is based on Eliott and Pfenning's work where success
   continuations are used to return results, and backtracking is
   automatic by sequencing.

   These are INTERNAL functions. Use the later function resolve_guard
   from outside.
*)

fun resolve_guard_internal (connection: Db.dbConnection) (g: guard) 
			   (su: Solution.subst) (isFinal: bool ref) 
			   (sc: Solution.subst -> unit): unit = 

    case g of
	gAtom (p, tl) => 
	let val tl' = subst_simplify_term_list su tl
	in
	    resolve_atom connection (p, tl') 
			 (fn (SOME su') => sc (su @ su')
			   | NONE => if Scenario.isFinal p tl'
				     then () else (isFinal := false)
 			 )
	end
      | gAnd gl => resolve_guard_conj_list_internal connection gl su isFinal sc 
      | gOr gl => resolve_guard_disj_list_internal connection gl su isFinal sc
      | gEx (vdl, g) =>
	resolve_guard_internal connection g su isFinal
			       (fn su' => sc (Solution.remove_subst su' (List.map #1 vdl)))
      | gTrue => sc su
      | gFalse => () (* backtrack *)
		  
and resolve_guard_conj_list_internal connection gl su isFinal sc =
    case gl of
	[] => sc su
      | g :: gl =>
	resolve_guard_internal connection g su isFinal
			       (fn su' => resolve_guard_conj_list_internal connection gl su' isFinal sc)

and resolve_guard_disj_list_internal connection gl su isFinal sc =
    case gl of
	[] => ()
      | g :: gl =>
	(resolve_guard_internal connection g su isFinal sc;
	 resolve_guard_disj_list_internal connection gl su isFinal sc
	)


(* Resolve a guard g by finding all its satisfying substitutions. For
   each satisfying substitution su, the success continuation sc is
   called once with su. The output parameter isFinal contains true iff
   there will be no more substitutions in future extensions, else it
   contains false.

   This function is a wrapper around resolve_guard_internal
*) 

fun resolve_guard (connection: Db.dbConnection) (g: guard) 
		  (isFinal: bool ref) 
		  (sc: Solution.subst -> unit): unit = 
    (isFinal := true;
     resolve_guard_internal connection g [] isFinal sc
    )


(* Check whether a fully ground non-SUBJ atom (p,tl) is true, false or
   unknown. Accordingly return tt, ff, or uu respectively.
 *)

datatype atomStatus = tt | ff | uu

local 
    exception trueFormula
in
fun check_ground_atom (conn: Db.dbConnection) (p, tl) =
    let val return = ref tt
    in
	resolve_atom conn (p, tl) 
		     (fn SOME [] => 
			 (* Found a satisfying instance (which must be
                            [] because tl is ground) so end the search
                          *)
			 (return := tt; raise trueFormula)
		       | SOME _ =>
			 (* Non-empty substitution; should not happen *)
			 (print ("In ReductionFn.check_ground_atom: " ^ 
				 "found non-empty satisfying susbtitution for atom: " ^ 
				 (print_atom (p,tl)) ^ "\n"); 
			  raise Impossible)
		       | NONE => if Scenario.isFinal p tl 
				 then return := ff
				 else return := uu
		     )	handle trueFormula => () ;
	(!return)
    end
end


(* Internal reduction function -- does a full simplication to the extent possible 
   * 3rd arg = true means instantiate quantifiers
   * 4th arg = true means lookup atoms

   Obviously, (3rd arg = true, 4th arg = false) does not make much sense
*)

fun reduce_formula_internal (conn: Db.dbConnection) (f:formula) (inst: bool) (lookup_atom: bool) =

    case f of
	fTrue _ => f
      | fFalse _ => f
      | fAtom (nopt, (p,tl)) =>
	if (lookup_atom) 
	then
	    (case lookup_decl declarations p of
		 SOME (declPred (_, predSubj, _, _)) => (* Subjective predicate *)
		 if (List.exists (fn (p', tl') => p' = p andalso tl' = tl) (!Scenario.subj_pred_list_true)) then fTrue nopt
		 else if (List.exists (fn (p', tl') => p' = p andalso tl' = tl) (!Scenario.subj_pred_list_false)) then fFalse nopt
		 else f
	       | SOME (declPred _) =>
		 (* DB/SQL/EVAL predicate *)
		 (case check_ground_atom conn (p, tl) of
		      tt => fTrue nopt
		    | ff => fFalse nopt
		    | uu => f
		 )
	       | _ =>
		 (* Not a predicate; should not happen by type checking *)
		 (print ("In function ReductionFn.reduce_formula_internal: Purported predicate " ^
			 (print_pred p) ^ "is not a declared predicate\n");
		  raise Impossible
		 )
	    )
	else f

      | fAnd (nopt, fl) =>
	(case reduce_formula_internal_conj_list conn fl inst lookup_atom of
	     NONE => fFalse nopt
	   | SOME [] => fTrue nopt
	   | SOME [f'] => f'
	   | SOME fl' => fAnd (nopt, fl')
	)

      | fOr (nopt, fl) =>
	(case reduce_formula_internal_disj_list conn fl inst lookup_atom of
	     NONE => fTrue nopt
	   | SOME [] => fFalse nopt
	   | SOME [f'] => f'
	   | SOME fl' => fOr (nopt, fl')
	)
	
      | formula as (fAll (nopt, vdl, exll, copt, g, f)) =>
	if (inst) then
	    let val reduced_list_ref = ref (SOME []) (* Will eventually contain reduced instances of f *)
		val sub_list = ref [] (* Will eventually contain the list of all substitutions considered *)
		val isFinal = ref true
		val vl = List.map #1 vdl 
		(* A continuation for resolve_guard that will create
                   instances of f and simplify them *)
		fun sc (su: Solution.subst) = 
		    (case !reduced_list_ref of
			 NONE => ()
		       | SOME fl' => 
			 let val tl = List.map (Option.valOf o (Solution.find_subst su)) vl
			 in 
			     case (List.exists (fn tl' => tl = tl') exll) of
				 true => () (* This substitution is excluded *)
			       | false =>
				 let (* val _ = print ("All: " ^ (Solution.print_subst su) ^ "\n") *)
				     val _ = sub_list := (tl :: (!sub_list))
				     val f_inst = reduce_formula_internal conn (subst_simplify_formula su f) inst lookup_atom
				 in
				     case f_inst of
					 fFalse _ => reduced_list_ref := NONE
				       | fTrue _ => () (* Do no include this *)
				       | _ => reduced_list_ref := SOME (f_inst :: fl')
				 end
			 end
		    )
		    
		val _ = resolve_guard conn g isFinal sc
			
		(* Reduced instances are now in reduced_ref_list, but in
                   reverse order. Add the residual formula, if needed. *)
		val _ = (case (!isFinal, !reduced_list_ref) of
			     (true, _) => () (* The substitutions are final *)
			   | (_, NONE) => () (* A falsity was already found *)
			   | (false, SOME fl') => 
			     let val exll' = (List.rev (!sub_list)) @ exll
			     in
				 reduced_list_ref := SOME (fAll (nopt, vdl, exll', copt, g, f) :: fl')
			     end)
			
		val instances = Option.map (List.rev) (!reduced_list_ref)
	    in
		case instances of
		    NONE => fFalse nopt
		  | SOME [] => fTrue nopt
		  | SOME [f'] => f'
		  | SOME fl' => fAnd(nopt, fl')
	    end
	else 
	    formula

      | formula as (fEx (nopt, vdl, exll, copt, g, f)) =>
	if (inst) then
	    let val reduced_list_ref = ref (SOME []) (* Will eventually contain reduced instances of f *)
		val sub_list = ref [] (* Will eventually contain the list of all substitutions considered *)
		val isFinal = ref true
		val vl = List.map #1 vdl 
		(* A continuation for resolve_guard that will create
		   instances of f and simplify them *)
		fun sc (su: Solution.subst) = 
		    (case !reduced_list_ref of
			 NONE => ()
		       | SOME fl' => 
			 let val tl = List.map (Option.valOf o (Solution.find_subst su)) vl
			 in 
			     case (List.exists (fn tl' => tl = tl') exll) of
				 true => () (* This substitution is excluded *)
			       | false =>
				 let (* val _ = print ("Ex: " ^ (Solution.print_subst su) ^ "\n") *)
				     val _ = sub_list := (tl :: (!sub_list))
				     val f_inst = reduce_formula_internal conn (subst_simplify_formula su f) inst lookup_atom
				 in
				     case f_inst of
					 fTrue _ => reduced_list_ref := NONE
				       | fFalse _ => () (* Do no include this *)
				       | _ => reduced_list_ref := SOME (f_inst :: fl')
				 end
			 end
		    )
		    
		val _ = resolve_guard conn g isFinal sc
			
		(* Reduced instances are now in reduced_ref_list, but in
                   reverse order. Add the residual formula, if needed. *)
		val _ = (case (!isFinal, !reduced_list_ref) of
			     (true, _) => () (* The substitutions are final *)
			   | (_, NONE) => () (* A truth was already found *)
			   | (false, SOME fl') => 
			     let val exll' = (List.rev (!sub_list)) @ exll
			     in
			     reduced_list_ref := SOME (fEx (nopt, vdl, exll', copt, g, f) :: fl')
			     end)
			
		val instances = Option.map (List.rev) (!reduced_list_ref)
	    in
		case instances of
		    NONE => fTrue nopt
		  | SOME [] => fFalse nopt
		  | SOME [f'] => f'
		  | SOME fl' => fOr(nopt, fl')
	    end
	else 
	    formula
	    
(* Reduce a list of formulas fl that are conjuncted. Return NONE if
   any formula in fl reduces to fFalse, else return SOME fl', where
   fl' are the reducts of fl, eliminating those that evaluate to fTrue
*)

and reduce_formula_internal_conj_list conn [] inst lookup_atom = SOME []
  | reduce_formula_internal_conj_list conn (f :: fl) inst lookup_atom =
    (case reduce_formula_internal conn f inst lookup_atom of
	 fTrue _ => reduce_formula_internal_conj_list conn fl inst lookup_atom 
       | fFalse _ => NONE
       | f' => Option.map 
		   (fn fl' => f' :: fl') 
		   (reduce_formula_internal_conj_list conn fl inst lookup_atom)
    )

(* Reduce a list of formulas fl that are disjuncted. Return NONE if
   any formula in fl reduces to fTrue, else return SOME fl', where fl'
   are the reducts of fl, eliminating those that evaluate to fFalse
*)

and reduce_formula_internal_disj_list conn [] inst lookup_atom = SOME []
  | reduce_formula_internal_disj_list conn (f :: fl) inst lookup_atom =
    (case reduce_formula_internal conn f inst lookup_atom of
	 fFalse _ => reduce_formula_internal_disj_list conn fl inst lookup_atom
       | fTrue _ => NONE
       | f' => Option.map 
		   (fn fl' => f' :: fl') 
		   (reduce_formula_internal_disj_list conn fl inst lookup_atom)
    )
    


(* Reduce fully *)
fun reduce_formula connection f = reduce_formula_internal connection f true true

(* Reduce, but do not instantiate quantifiers *)
fun reduce_formula_no_inst connection f = reduce_formula_internal connection f false true
								       
(* Reduce, but contract only trues and falses. Here the database
   cannot be used, so a dummyConnection works *)
fun reduce_simplify f = reduce_formula_internal Db.dummyConnection f false false
				
(* If the top level connective is a quantifier, instantiate it, else do nothing *)
fun instantiate_formula conn f = 
    let fun make_instances (nopt, vdl, exll, copt, g, f) =
	    let val inst_list_ref = ref [] (* Will eventually contain reduced instances of f *)
		val sub_list = ref [] (* Will eventually contain the list of all substitutions considered *)
		val isFinal = ref true
		val vl = List.map #1 vdl 
		(* A continuation for resolve_guard that will create instances of f *)
		fun sc (su: Solution.subst) = 
		    let val tl = List.map (Option.valOf o (Solution.find_subst su)) vl
		    in 
			case (List.exists (fn tl' => tl = tl') exll) of
			    true => () (* This substitution is excluded *)
			  | false =>
			    let (* val _ = print ("All: " ^ (Solution.print_subst su) ^ "\n") *)
				val _ = sub_list := (tl :: (!sub_list))
				val f_inst = subst_simplify_formula su f
			    in
				inst_list_ref := f_inst :: (!inst_list_ref)
			    end
		    end
		    
		    
		val _ = resolve_guard conn g isFinal sc
			
		(* Reduced instances are now in inst_ref_list, but in
                   reverse order. Add the residual formula, if
                   needed. 
                 *)
		val _ = (case (!isFinal) of
			     true => () (* The substitutions are final *)
			   | false => 
			     let val exll' = (List.rev (!sub_list)) @ exll
			     in
				 inst_list_ref := fAll (nopt, vdl, exll', copt, g, f) :: (!inst_list_ref)
			     end
			)
	    in
		List.rev (!inst_list_ref)
	    end

    in
	case f of
	    fAll (nopt, vdl, exll, copt, g, f) =>
	    (case (make_instances (nopt, vdl, exll, copt, g, f)) of
		 [] => fTrue nopt
	       | [f'] => f'
	       | fl' => fAnd(nopt, fl')
	    )
	  | fEx (nopt, vdl, exll, copt, g, f) =>
	    (case (make_instances (nopt, vdl, exll, copt, g, f)) of
		 [] => fFalse nopt
	       | [f'] => f'
	       | fl' => fOr(nopt, fl')
	    )
	  | _ => f
    end


end

end
