(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


*)

(*************************************************************************)

(*************************************************************************)

(* A signature that describes the scenario-specific interface. This
   signature includes the following:

   - An interface to the system database that must conform to the
     signature DB

   - The name of the system database to use

   - A function for simplifying EVAL constructors with value arguments
     (The function must also cover SQL constructors). The result must
     be a value.

   - A function for simplifying EVAL predicates. In this case, a
     success continuation must be called with every valid substitution
     for the output positions. This function must also work on SQL
     predicates.

   - Checking whether there will be other satisfying substitutions for
     a DB predicate in future extensions.
*)
   
signature SCENARIO = sig
    
    structure Db: DB
    (* An interface to the system database *)

    val db_name: Db.dbName
    (* Name of the system database to use *)


    val declarations: Declaration.decls
    (* The signature of the scenario *)

    val eval_term: Fol.const -> Fol.term list -> Fol.term
    (* eval_term c tl should return the simplication of tApp(c, tl).
       It is guaranteed that c is a constructor of type EVAL or SQL
       and that tl contains only values of the correct argument
       sorts. The result must be a value of the correct result
       sort. The latter fact may be checked by the calling program.
     *)

    val resolve_eval_pred: Db.dbConnection -> Fol.pred -> Fol.term list 
			   -> (Solution.solution -> 'a) 
			   -> 'a
    (* In (resolve_eval_pred p tl sc), it can be assumed that p is a
       predicate of type EVAL or SQL and that all input positions of p
       in tl are ground. For each satisfying solution tl', the success
       continuation sc must be invoked once with an argument of form
       (NEXT tl'). Finally, as the last statement, the success
       continuation must also be invoked once with NOMORE. Note that a
       satisfying solution is the entire list of terms satisfying the
       predicate, including those in input positions (these must
       simply be copied to the solution from tl). The function does
       not have to check the output positions at all (the caller must
       make the matches), but it may choose to do so for optimization.
     *)


    val subj_pred_list_true: (Fol.pred * Fol.term list) list ref
    (* List of subj ground predicates that have been marked true by a
       human being.
     *)

    val subj_pred_list_false: (Fol.pred * Fol.term list) list ref
    (* List of subj ground predicates that have been marked false by a
       human being.
     *)

    val isFinal: Fol.pred -> Fol.term list -> bool
    (* hasNoMore p tl: For a non-SUBJ predicate p, specify if at the
       end of query (p tl) there will be further solutions or not in
       further extensions of the log. Return true if there will be no
       more solutions in any extension and false otherwise. Note that
       the first argument is guaranteed to be a non-SUBJ predicate and
       all input arguments in tl are guaranteed to be ground and of
       the correct sorts.

       If the databases are past-complete, one simple way to implement
       this function is to simply look at the time argument. If it is
       ground and less than the current time, return true, else return
       false.
    *)

 end
