(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)


(* Description of solutions to predicates and functions for
   matching. We keep matters simple and define a substitution as a
   list of (var, value) pairs.
*)


structure Solution = struct

local 
    open Fol

in

type subst = (var * term) list
(* Invariant: Each term in a subst must be a value *)

exception SubstNotFound of var

fun find_subst (su: subst) (x: var) = 
    case List.find (fn (y, _) => x = y) su of
	NONE => NONE
      | SOME (_, v) => SOME v


(* Remove a list of variables from a substitution *)

fun remove_subst (su: subst) ([]: var list) = su
  | remove_subst su (v :: vl) = 
    let val su' = List.filter (fn (v', _) => v <> v') su
    in
	remove_subst su' vl
    end


fun print_subst (su: subst) =
    List.foldr (fn ((v, t), s) => (Fol.print_var v) ^ 
				   ": " ^ 
				   (Fol.print_term t) ^
				   "; " ^ s) "" su


(* Apply a substitution su to a syntactic entity
*)

fun subst_term (su: subst) (t: term) = 
    case t of
	tVar x =>
	(case find_subst su x of
	     NONE => tVar x
	   | SOME v => v
	)
      |	tString _ => t
      | tInt _ => t
      | tDate _ => t
      | tApp (c, tl) => tApp (c, subst_term_list su tl)

and subst_term_list su tl = List.map (subst_term su) tl


fun subst_atom su ((p, tl): atom) = 
    (p, subst_term_list su tl)

fun subst_guard su (g: guard) = 
    case g of
	gAtom a => gAtom (subst_atom su a)
      | gAnd gl => gAnd (subst_guard_list su gl)
      | gOr gl => gOr (subst_guard_list su gl)
      | gEx (vdl, g) => gEx (vdl, subst_guard su g)
      | gTrue => gTrue
      | gFalse => gFalse

and subst_guard_list su gl = List.map (subst_guard su) gl

fun subst_formula su (f: formula) = 
    case f of
	fAtom (nopt, a) => fAtom (nopt, subst_atom su a)
      | fAnd (nopt, fl) => fAnd (nopt, subst_formula_list su fl)
      | fOr (nopt, fl) => fOr (nopt, subst_formula_list su fl)
      | fEx (nopt, vdl, exll, copt, g, f) =>
	fEx (nopt, vdl, List.map (subst_term_list su) exll, copt,
	     subst_guard su g, subst_formula su f)
      | fAll (nopt, vdl, exll, copt, g, f) =>
	fAll (nopt, vdl, List.map (subst_term_list su) exll, copt,
	      subst_guard su g, subst_formula su f)
      | fTrue nopt => fTrue nopt
      | fFalse nopt => fFalse nopt

and subst_formula_list su fl = List.map (subst_formula su) fl

(* Matching -- match a term to another. The second one is guaranteed
   to be ground. The function is also given an existing substitution
   for the variables in the first term. This may initially be
   empty. The output of the function is either SOME subst, if subst is
   the (unique) matching substitution, or NONE if the terms cannot be
   matched. If the output is SOME subst, then subst will be an
   extension of the input substitution.

   It is assumed (without checking) that all constructors appearing in
   either term are of type SYM. 
*)

fun match_terms su t1 t2 = 
    case (t1, t2) of
	(tVar v, _) =>
	(case find_subst su v of
	     NONE => SOME ((v, t2) :: su)
	   | SOME t' => if (t' = t2) 
			then SOME su 
			else NONE
	)

      | (tString s1, tString s2) => if (s1 = s2) 
				    then SOME su
				    else NONE
      | (tInt i1, tInt i2) => if (i1 = i2) 
			      then SOME su
			      else NONE
      | (tDate d1, tDate d2) => if (d1 = d2)
				then SOME su 
				else NONE

      | (tApp (c1, tl1), tApp (c2, tl2)) =>
	if (c1 = c2) 
	then match_term_lists su tl1 tl2
	else NONE

      | _ => NONE

and match_term_lists su [] [] = SOME su
  | match_term_lists su (t1 :: tl1) (t2 :: tl2) =
    (case match_terms su t1 t2 of
	 NONE => NONE
       | SOME su' => match_term_lists su' tl1 tl2
    )
  | match_term_lists su _ _ = NONE




(* A datatype used to return a solution to a predicate. NOMORE means
   that there are no more substitutions at present. 
*)

datatype solution = NOMORE
		  | NEXT of term list

end

end
