(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


*)

(*************************************************************************)

(*************************************************************************)


(* Abstract definition and implementation of SQL SELECT queries *)

functor SQLFn (Db: DB) = struct

structure Db = Db

type colname = string * string * string Option.option
               (* tablename * colname * as_name 
		  T.name as x, T'.ssn as y, etc. *)

datatype sql_term = folTerm of Fol.term
		  | colName of string * string (* tablename.columname *)
		  | identifier of string (* identifier that names a column *)

datatype sql = Select of colname list * from list * constraint list
	 
     and from = Table of string (* Name of table *)
	      | InnerSQL of sql (* Inner SQL query *)

     and constraint = Ceq of sql_term * sql_term


(*************************************************************************)

(*************************************************************************)

fun print_colname (tn, cn, NONE) =
    tn ^ "." ^ cn
  | print_colname (tn, cn, SOME asn) = 
    tn ^ "." ^ cn ^ " as " ^ asn

and print_colname_list [] = ""
  | print_colname_list [c] = print_colname c
  | print_colname_list (c :: cl) =
    (print_colname c) ^ ", " ^ (print_colname_list cl)

fun print_sql_term (folTerm t) = "'" ^ (Db.escape (Fol.print_term t)) ^ "'"
  | print_sql_term (colName (t, c)) = t ^ "." ^ c
  | print_sql_term (identifier s) = s

fun print_constraint (Ceq (t1,t2)) = 
    "(" ^ (print_sql_term t1) ^ " = " ^ (print_sql_term t2) ^ ")"

and print_constraint_list [] = ""  (* Should not happen *)
  | print_constraint_list [c] = print_constraint c
  | print_constraint_list (c :: cl) =
    (print_constraint c) ^ " AND " ^ (print_constraint_list cl)

fun print_from (Table s) = s
  | print_from (InnerSQL sql) = "(" ^ (print_sql sql) ^ ")"

and print_from_list [] = ""
  | print_from_list [f] = print_from f
  | print_from_list (f :: fl) = 
    (print_from f) ^ ", " ^ (print_from_list fl)

and print_sql (Select (cnl, fl, [])) =
    "SELECT " ^ (print_colname_list cnl) ^ 
    " FROM " ^ (print_from_list fl) 
  | print_sql (Select (cnl, fl, cl)) =
    "SELECT " ^ (print_colname_list cnl) ^ 
    " FROM " ^ (print_from_list fl) ^
    " WHERE " ^ (print_constraint_list cl)



(*************************************************************************)

(*************************************************************************)


(* Take a DB atom (p, tl) possibly with free variables and find
   satisfying terms lists from the database that match the input
   positions, which can be assumed ground. The function should call
   its continuation once with each satisfying term list. After all
   calls with satisfying term lists, the success continuation is to be
   called one last time with NOMORE.

   Also, the function does not have to match the output positions to
   terms obtained from the database, although for optimization its
   current implementation will make matches in output positions that
   are completely ground. *)

exception CannotPrepareQuery of string
exception UnexpectedUnequalLengths
exception ErrorRunningDbQuery of string
exception UnparseableTerm of string
exception BadDeclaration of string

fun resolve_db_pred (ds: Declaration.decls) (connection: Db.dbConnection) 
		    (p: Fol.pred) (tl: Fol.term list) 
		    (sc: Solution.solution -> 'a): 'a =
    let
	(* Name of table for predicate p, its column names and the
           types of arguments.  Note that modes are irrelevant during
           the execution; they are merely static *)
	val (tablename, colnames, sorts) = 
	    case (Declaration.lookup_decl ds p) of
		SOME (Declaration.declPred (sml, Declaration.predDB, NONE, SOME (tn, cols))) => 
		(tn, cols, List.map #1 sml)
	      | _ => 
		(* This should not happen unless the declarations are
	           not checked or this function is called on a non-DB
	           predicate *)
		(print ("Bad/unfound declaration for purported DB predicate: "
			^ (Fol.print_pred p) ^ "\n");
		 raise BadDeclaration p
		)

	(* Convert column list to SQL format *)
	val col_list = List.map (fn cn => (tablename, cn, NONE)) colnames

	(* Construct constraints for the WHERE clause of the SQL
	   query. The constraints simply equate grounded arguments to
	   the corresponding columns. This implies that for a
	   mode-checked input, at least all columns in input positions
	   will be constrained.
         *)
	val constraints = 
	    let fun collect_constraints ([]: Fol.term list) ([]: string list) = []
		  | collect_constraints (t :: tl) (s :: sl) = 
		    if (Fol.isClosed t) 
		    then (Ceq (colName (tablename, s), folTerm t))
			 :: (collect_constraints tl sl)
		    else collect_constraints tl sl
		  | collect_constraints _ _ = 
		    (* Should not happen by type checking *)
		    (print ("Unexpected unequal lengths of args: " ^
			    (Fol.print_term_list tl) ^ " and column names: " ^
			    (String.concatWith " " colnames) ^ "\n");
		     raise UnexpectedUnequalLengths)
	    in
		collect_constraints tl colnames
	    end		

	(* The SQL query to run *)
	val sql = Select (col_list, [Table tablename], constraints) 

	(* The string representation of the query *)
	val sql_string = print_sql sql

	(* Abstract, prepared query *)
	val query = case (Db.newQuery connection sql_string) of
			NONE => (print ("Unable to prepare query: " ^ 
					(sql_string) ^ "\n");
				 raise CannotPrepareQuery sql_string)
		      | SOME q => q

	(* The main function that loops through the results of the query *)
	fun loop () =
	    case Db.stepQuery query of
		Db.ERROR => 
		(Db.destroyQuery query;
		 raise ErrorRunningDbQuery sql_string)
	      | Db.DONE => 
		(* There are no more results. *)
		(Db.destroyQuery query;
		 sc Solution.NOMORE)
	      | Db.DATA =>
		let 
		    (* Get the current row of query *)
		    val stringlist = Db.getData query

		    (* Parse the returned strings to terms *)
		    val termlist = List.map 
				       (fn s => 
					   case (FolParser.parse_term s) of
					       SOME t => t
					     | NONE => (print ("Unparseable term: " ^ s ^ 
							       " obtained from database\n");
							raise UnparseableTerm s
						       )
				       ) stringlist

		    (* Check that the returned terms are values of expected sorts *)
		    (* This is a DYNAMIC CHECK that can be eliminated,
		       if it can be guaranteed that sorts are checked
		       whenever data is written to the database *)
		  
		    val _ = Tc.tc_term_value_list ds sorts termlist 

		    (* Call success continuation *)
		    val _ = sc (Solution.NEXT termlist)
		in
		    loop ()
		end

    in
	loop ()
    end


end
