(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* An implementation of SQLite version 3 databases. *)

structure SQLiteDb :> DB =
struct
 
type dbName = string 
(* SQLite databases are identified by file names *)

type dbConnection = MLton.Pointer.t
(* This is actually the C type sqlite3* *)
		    
type query = MLton.Pointer.t 
(* This is actually the C type sqlite_stmt* *)

datatype queryresult = DATA | DONE | ERROR

exception SQLiteCInterfaceException

(* An internal structure that wraps around native C SQLite3 functions *)

structure Internal = struct

val smli_sqlite3_open = _import "smli_sqlite3_open": string -> MLton.Pointer.t;
(* Takes a string representing a SQLite3 database name and returns a
   pointer (dbConnection). The connection is readOnly. The pointer is
   null iff there was an error.
*)


val smli_sqlite3_close = _import "smli_sqlite3_close": MLton.Pointer.t -> unit;
(* Takes a dbConnection pointer and close the connection *)

val smli_sqlite3_prepare = _import "smli_sqlite3_prepare": MLton.Pointer.t * string -> MLton.Pointer.t;
(* Takes a dbConnection pointer and a SQL query and compiles it to a
   SQLite3 sqlite_stmt pointer
*)

val smli_sqlite3_step = _import "smli_sqlite3_step": MLton.Pointer.t -> int;
(* Take a sqlite_stmt pointer and step it once. The output is coded as follows:
   0: new data is available (SQLITE_ROW)
   1: query is done executing (SQLITE_DONE)
   2: error occurred (SQLITE_something_else)
*)

val smli_sqlite3_column_count = _import "smli_sqlite3_column_count": MLton.Pointer.t -> int;
(* Takes a sqlite_stmt and returns the number of columns each of its steps
   will contain. This does not execute the statement. 
*)

val smli_sqlite3_column_bytes = _import "smli_sqlite3_column_bytes": MLton.Pointer.t * int -> int;
(* Returns the number of bytes (not counting the \0 terminator) in the
   nth column of a returned row. Result is zero iff row is null.
*)

val smli_sqlite3_column_text = _import "smli_sqlite3_column_text": MLton.Pointer.t * int * char array -> unit;
(* Copy the string in the nth column of the returned row into the
   third argument.  The char array must have one more byte than the
   integer returned by smli_sqlite3_column_bytes()
*)

val smli_sqlite3_reset = _import "smli_sqlite3_reset": MLton.Pointer.t -> unit;
(* Reset a query to re-run it *)

val smli_sqlite3_finalize = _import "smli_sqlite3_finalize": MLton.Pointer.t -> unit;
(* Destroy a sqlite_stmt object *)

end (* structure Internal *)



(* For the meanings of these functions, see the file db.sig *)

fun escape (s: string) = 
    String.translate (fn #"'" => "''"
		       | c => String.implode [c]) s

fun newConnection (name: dbName): dbConnection Option.option =
    let val p = Internal.smli_sqlite3_open (name ^ "\000")
    in
	if p = MLton.Pointer.null
	then NONE
	else SOME p
    end


val dummyConnection = MLton.Pointer.null

fun closeConnection (c: dbConnection) = 
    Internal.smli_sqlite3_close c

fun newQuery (c: dbConnection) (sql: string): query Option.option =
    let val q = Internal.smli_sqlite3_prepare (c, sql ^ "\000")
    in
	if (q = MLton.Pointer.null)
	then NONE
	else SOME q
    end


fun stepQuery (q: query) =
    case (Internal.smli_sqlite3_step q) of
	0 => DATA
      | 1 => DONE
      | 2 => ERROR
      | _ => raise SQLiteCInterfaceException


fun getData (q: query): string list =
    let val nCols = Internal.smli_sqlite3_column_count q

        (* Extract all columns after n. sl is the list of previous
           columns in reverse order *)
	fun helper n (sl: string list) = 
	    if (n = nCols) then (List.rev sl)
	    else 
		let val currLen = Internal.smli_sqlite3_column_bytes (q, n)
		    val buf = Array.array (currLen + 1, #"\000")
		    val _ = Internal.smli_sqlite3_column_text (q, n, buf)
		    val str: string = CharVector.tabulate (currLen, fn i => Array.sub(buf,i)) 
		in
		    helper (n+1) (str :: sl)
		end
    in 
	helper 0 []
    end

fun resetQuery (q: query): unit = 
    Internal.smli_sqlite3_reset q

fun destroyQuery (q: query): unit = 
    Internal.smli_sqlite3_finalize q


end
