(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*************************************************************************)

(*************************************************************************)

(* Sort reconstruction and type-checking for FOL. We infer the sorts
   of bound variables if they have not been given by the user
   already. The idea is as follows:

   - If the user specifies the sort of a variable then we only check
     that this specified sort equal to the sort expected at each
     occurrence of the variable.

   - If the user does not specify the sort of a variable, then we
     assume its sort is unknown till we reach an occurrence, at which
     point that occurrence defines the sort.
*)

signature TC = sig
    (* Perform sort inference and return the new formula with all
       sorts in bindings resolved; print error and raise exception
       otherwise *)
    val tc_formula: Declaration.decls -> Fol.formula -> Fol.formula

    (* Check that a term is a value of a particular sort. Print error
       and raise exception on error. Return () on success *)
    val tc_term_value: Declaration.decls -> Declaration.sort -> Fol.term -> unit
    val tc_term_value_list: Declaration.decls -> Declaration.sort list -> Fol.term list -> unit
end


structure Tc: TC = struct

exception SortClash of Fol.term
exception VarNotFound of Fol.var
exception UndeclaredConstructor of Fol.const
exception UndeclaredPredicate of Fol.pred
exception TooFewTerms of Declaration.sort list
exception TooManyTerms of Fol.term list
			       
local
    open Declaration
    open Fol

    datatype curr_sort = unknown
		       (* Sort is not known *)
		       | inferred of sort 
                       (* Inferred sort *)
		       | user of sort
                       (* user provided sort *)

    type sortmap = (var * curr_sort ref) list

    local 
	(* Find the reference cell containing the sort of v. *)
	fun lookup_ref (sm: sortmap) (v: var) = 
	    Option.map (fn (_,y) => y) (List.find (fn (v', _) => v' = v) sm)

    in
    (* Find the sort of v in the sm *)
    fun lookup_sort sm v = Option.map (fn l => !l) (lookup_ref sm v)
			 
    (* Update the sort of v to s. If v is not in the domain, do nothing *)
    fun update_sort sm v s = case (lookup_ref sm v) of
				 NONE => ()  (* This case should not happen *)
			       | SOME l => (l := s)

    end

    (* Check that term t has sort ex in declarations ds and sort map
    sm for variables. Return () on success and print error, raise
    exception otherwise. *)

    fun infer_term (ds: decls) (sm: sortmap) 
		   (ex: sort) (t: term) =
                   (* ex is the expected sort of t *)
	case t of
	    tVar v =>
	    (case lookup_sort sm v of
		 NONE => 
		 (* This case should not happen when we start from a
		    formula obtained from preproc_formula because if v
		    is not in sm, then v is a constructor, not a
		    variable *) 
		 raise VarNotFound v
		       
	       | SOME (user s) => 
		 (* The user gave us sort s for variable v, so all we
                    have to check is that s equal to ex *)
		 if (equal_sort s ex) then ()
		 else (print ("Sort clash on variable: " ^ (print_var v) ^ 
			      ". User provided sort: " ^ (print_sort s) ^ 
			      ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		       raise SortClash t)

	       | SOME (inferred s) => 
		 (* We previously inferred sort s for variable v, so
                    all we have to check is that s is equal to ex
                 *)
		 if (equal_sort s ex) then ()
		 else (print ("Sort clash on variable: " ^ (print_var v) ^ 
			      ". Inferred sort: " ^ (print_sort s) ^ 
			      ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		       raise SortClash t)

	       | SOME unknown => 
		 (* Sort of v is unknown. Set it to ex now *)
		 update_sort sm v (inferred ex)
	    )

	  | tString _ => 
	    if (equal_sort sortString ex) then ()
	    else (print ("Sort clash on term: " ^ (print_term t) ^ 
			 ". It's sort: " ^
			 (print_sort sortString) ^ 
			 ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		  raise SortClash t)

	  | tInt _ => 
	    if (equal_sort sortInt ex) then ()
	    else (print ("Sort clash on term: " ^ (print_term t) ^ 
			 ". It's sort: " ^ 
			 (print_sort sortInt) ^ 
			 ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		  raise SortClash t)

	  | tDate _ => 
	    if (equal_sort sortDate ex) then ()
	    else (print ("Sort clash on term: " ^ (print_term t) ^ 
			 ". It's sort: " ^ 
			 (print_sort sortDate) ^ 
			 ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		  raise SortClash t)

	  | tApp (c, tl) => 
	    (case lookup_decl ds c of  
		 NONE => 
		 (print ("Undeclared constructor: " ^ (print_const c) ^ "\n");
		  raise UndeclaredConstructor c)

	       | SOME (declConst (sl, s, _, _)) =>
		 if (equal_sort s ex) 
		 then infer_term_list ds sm sl tl
		 else (print ("Sort clash on term: " ^ (print_term t) ^ 
			      ". Declared sort from signature: " ^ 
			      (print_sort s) ^ 
			      ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		       raise SortClash t)

	       | SOME _ => (* c is not a constructor *)
		 (print ("Non-constructor: " ^ (print_const c) ^ " used as constructor\n");
		  raise UndeclaredConstructor c)
	    )

    and infer_term_list ds sm sl tl = 
	case (sl, tl) of
	    ([], []) => ()
	  | (s :: _, []) => (print ("Missing expected argument term of sort " ^ 
				     (print_sort s) ^ "\n");
			      raise TooFewTerms sl)
	  | ([] , _ :: _) => (print ("Too many argument terms starting at: " ^ 
				     (print_term_list tl) ^ "\n");
			      raise TooManyTerms tl)
	     
	  | (s :: sl', t :: tl') => 
	    (infer_term ds sm s t; infer_term_list ds sm sl' tl')


    fun infer_atom ds sm ((p, tl): atom) =
	    (case lookup_decl ds p of  
		 NONE => 
		 (print ("Undeclared predicate symbol: " ^ (print_pred p) ^ "\n");
		  raise UndeclaredPredicate p)

	       | SOME (declPred (sml, _, _, _)) =>  (* sml: (sort * mode) list *)
		 infer_term_list ds sm (List.map (#1) sml) tl

	       | SOME _ => (* p is not a predicate symbol *)
		 (print ("Non-predicate: " ^ (print_pred p) ^ " used as predicate\n");
		  raise UndeclaredPredicate p)
	    )
	
    (* Functions to convert between (var * sort option) list and sortmap *)

    fun bound_list_to_sortmap (vdl: (var * sort option) list): sortmap =
	List.map (fn (v, NONE) => (v, ref (unknown))
		   | (v, SOME s) => (v, ref (user s))) vdl

    fun sortmap_to_bound_list (sm: sortmap): (var * sort option) list =
	List.map (fn (v, ref (user s)) => (v, SOME s)
		   | (v, ref (inferred s)) => (v, SOME s)
		   | (v, ref unknown) => (v, SOME sortAny)) sm


    (* Check that all sorts in a binding list are declared. Return
       anything on success if true, raise exception otherwise *)

    exception UndeclaredSort of sort
			
    fun check_sorts_declared ds vdl = 
	List.map (fn (_, NONE) => ()
		   | (_, SOME (s' as (sortOther s))) =>
		     (case lookup_decl ds s of
			  NONE => (print ("Sort " ^ (print_sort s') ^ 
					  " undeclared\n");
				   raise UndeclaredSort s')
			| SOME declSort => ()
			| SOME _ => (print ("Sort " ^ (print_sort s') ^ 
					    " is not a sort\n");
				     raise UndeclaredSort s')
		     )
		   | (_, SOME _) => ()
		 ) vdl
			 
		     
	
    (* Sort inference on guard g. Returns a new guard with correctly
       inferred sorts on success and raises exception otherwise *)

    fun infer_guard ds sm (g: guard): guard =
	case g of
	    gAtom a => (infer_atom ds sm a; g)
	  | gAnd gl => gAnd (infer_guard_list ds sm gl)
	  | gOr gl => gOr (infer_guard_list ds sm gl)
	  | gEx (vdl, g') => 
	    let val _ = check_sorts_declared ds vdl
		val sm' = bound_list_to_sortmap vdl
		val g'' = infer_guard ds (sm' @ sm) g'
                (* The above will modify the references in sm' *)
		val vdl' = sortmap_to_bound_list sm'
	    in
		gEx (vdl', g'')
	    end
	  | gTrue => gTrue
	  | gFalse => gFalse

    and infer_guard_list ds sm gl = List.map (infer_guard ds sm) gl

    fun infer_formula ds sm (f: formula): formula =
	case f of
	    fAtom (_, a) => (infer_atom ds sm a; f)
	  | fAnd (n_opt, fl) => fAnd (n_opt, infer_formula_list ds sm fl)
	  | fOr (n_opt, fl) => fOr (n_opt, infer_formula_list ds sm fl)
	  | fEx (n_opt, vdl, exl, comp_opt, g, f') =>
	    let val _ = check_sorts_declared ds vdl
		val sm' = bound_list_to_sortmap vdl
		val g' = infer_guard ds (sm' @ sm) g
		val f'' = infer_formula ds (sm' @ sm) f'
		val vdl' = sortmap_to_bound_list sm'
		val sl = List.map (fn (_, SOME s) => s
				    | (_, NONE) => sortAny) vdl'
		val _ = List.map (infer_term_list ds sm sl) exl
	    in 
		fEx (n_opt, vdl', exl, comp_opt, g', f'')
	    end
	  | fAll (n_opt, vdl, exl, comp_opt, g, f') =>
	    let val _ = check_sorts_declared ds vdl
		val sm' = bound_list_to_sortmap vdl
		val g' = infer_guard ds (sm' @ sm) g
		val f'' = infer_formula ds (sm' @ sm) f'
		val vdl' = sortmap_to_bound_list sm'
		val sl = List.map (fn (_, SOME s) => s
				    | (_, NONE) => sortAny) vdl'
		val _ = List.map (infer_term_list ds sm sl) exl
	    in 
		fAll (n_opt, vdl', exl, comp_opt, g', f'')
	    end
	  | fTrue _ => f
	  | fFalse _ => f	

    and infer_formula_list ds sm fl = List.map (infer_formula ds sm) fl


in

fun tc_formula ds f = infer_formula ds [] f


(*********************************************************************************)

(*********************************************************************************)

(* Functions for checking that terms are values of given sorts *)

exception VarInValue of var
exception NonSymInValue of const

(* Check that term t is a value of sort ex in declarations ds. Return
   () on success and print error, raise exception otherwise. 
*)

fun tc_term_value (ds: decls)  (ex: sort) (t: term) =
    (* ex is the expected sort of t *)
    case t of
	tVar v =>
	(print ("Variable " ^ (print_var v) ^ " found in purported value\n");
	 raise VarInValue v
	)
      | tString _ => 
	if (equal_sort sortString ex) then ()
	else (print ("Sort clash on term: " ^ (print_term t) ^ 
		     ". It's sort: " ^
		     (print_sort sortString) ^ 
		     ". Expected sort: " ^ (print_sort ex) ^ ".\n");
	      raise SortClash t)

      | tInt _ => 
	if (equal_sort sortInt ex) then ()
	else (print ("Sort clash on term: " ^ (print_term t) ^ 
		     ". It's sort: " ^ 
		     (print_sort sortInt) ^ 
		     ". Expected sort: " ^ (print_sort ex) ^ ".\n");
	      raise SortClash t)
	     
      | tDate _ => 
	if (equal_sort sortDate ex) then ()
	else (print ("Sort clash on term: " ^ (print_term t) ^ 
		     ". It's sort: " ^ 
		     (print_sort sortDate) ^ 
		     ". Expected sort: " ^ (print_sort ex) ^ ".\n");
	      raise SortClash t)
	     
      | tApp (c, tl) => 
	(case lookup_decl ds c of  
	     NONE => 
	     (print ("Undeclared constructor: " ^ (print_const c) ^ "\n");
	      raise UndeclaredConstructor c)
	     
	   | SOME (declConst (sl, s, constSym, _)) => (* c is a symbolic constructor *)
	     if (equal_sort s ex) 
	     then tc_term_value_list ds sl tl
	     else (print ("Sort clash on term: " ^ (print_term t) ^ 
			  ". Declared sort from signature: " ^ 
			  (print_sort s) ^ 
			  ". Expected sort: " ^ (print_sort ex) ^ ".\n");
		       raise SortClash t)
		  
	   | SOME (declConst (sl, s, _, _)) => (* c is a non-symbolic constructor *)
	     (print ("Non-sym constructor" ^ (print_const c) ^ " appears in value\n");
	      raise NonSymInValue c)
	   | SOME _ => (* c is not a constructor *)
	     (print ("Non-constructor: " ^ (print_const c) ^ " used as constructor\n");
	      raise UndeclaredConstructor c)
	)
	
and tc_term_value_list ds sl tl = 
    case (sl, tl) of
	([], []) => ()
      | (s :: _, []) => (print ("Missing expected argument term of sort " ^ 
				(print_sort s) ^ "\n");
			 raise TooFewTerms sl)
      | ([] , _ :: _) => (print ("Too many argument terms starting at: " ^ 
				 (print_term_list tl) ^ "\n");
			  raise TooManyTerms tl)
			 
      | (s :: sl', t :: tl') => 
	(tc_term_value ds s t; tc_term_value_list ds sl' tl')



end

end

